"""minos.common.queries module."""

from __future__ import (
    annotations,
)

import re
from abc import (
    ABC,
    abstractmethod,
)
from functools import (
    partial,
)
from operator import (
    attrgetter,
)
from typing import (
    Any,
    Callable,
    Iterable,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    Model,
)


class _Condition(ABC):
    def evaluate(self, value: Model) -> bool:
        """Evaluate if given value satisfied this condition.

        :param value: The value to be evaluated.
        :return: A boolean value.
        """
        return self._evaluate(value)

    @abstractmethod
    def _evaluate(self, value: Model) -> bool:
        pass

    def __eq__(self, other) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable[Any]:
        yield from tuple()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(map(repr, self))})"


class _TrueCondition(_Condition):
    def _evaluate(self, value: Model) -> bool:
        return True


class _FalseCondition(_Condition):
    def _evaluate(self, value: Model) -> bool:
        return False


class _ComposedCondition(_Condition, ABC):
    def __init__(self, *parts: _Condition):
        self.parts = tuple(parts)

    def __iter__(self):
        yield from self.parts


class _AndCondition(_ComposedCondition):
    def _evaluate(self, value: Model) -> bool:
        return all(c.evaluate(value) for c in self.parts)


class _OrCondition(_ComposedCondition):
    def _evaluate(self, value: Model) -> bool:
        return any(c.evaluate(value) for c in self.parts)


class _NotCondition(_Condition):
    def __init__(self, inner: _Condition):
        self.inner = inner

    def _evaluate(self, value: Model) -> bool:
        return not self.inner.evaluate(value)

    def __iter__(self) -> Iterable[Any]:
        yield from (self.inner,)


class _SimpleCondition(_Condition, ABC):
    def __init__(self, field: str, parameter: Any):
        self.field = field
        self.parameter = parameter

    def __iter__(self) -> Iterable[Any]:
        yield from (
            self.field,
            self.parameter,
        )

    @property
    def _get_field(self) -> Callable[[Any], Any]:
        return attrgetter(self.field)


class _LowerCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self._get_field(value) < self.parameter


class _LowerEqualCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self._get_field(value) <= self.parameter


class _GreaterCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self._get_field(value) > self.parameter


class _GreaterEqualCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self._get_field(value) >= self.parameter


class _EqualCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self._get_field(value) == self.parameter


class _NotEqualCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self._get_field(value) != self.parameter


class _InCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self._get_field(value) in self.parameter


class _ContainsCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return self.parameter in self._get_field(value)


class _LikeCondition(_SimpleCondition):
    def _evaluate(self, value: Model) -> bool:
        return bool(self._pattern.fullmatch(self._get_field(value)))

    @cached_property
    def _pattern(self):
        return re.compile(self.parameter.replace("%", ".*").replace("_", "."))


_TRUE_CONDITION = _TrueCondition()
_FALSE_CONDITION = _FalseCondition()


class Condition:
    """Condition class.

    This class provides the way to create filtering conditions for ``Model`` instances based on the following operators:

    * `TRUE`: Always evaluates as `True`.
    * `FALSE`: Always evaluates as `True`.
    * `AND`: Only evaluates as `True` if all the given conditions are evaluated as `True`.
    * `OR`: Evaluates as `True` if at least one of the given conditions are evaluated as `True`.
    * `NOT`: Evaluates as `True` only if the inner condition is evaluated as `False`.
    * `LOWER`: Evaluates as `True` only if the field of the given model is lower (<) than the parameter.
    * `LOWER_EQUAL`: Evaluates as `True` only if the field of the given model is lower or equal (<=) to the parameter.
    * `GREATER`: Evaluates as `True` only if the field of the given model is greater (>) than the parameter.
    * `GREATER_EQUAL`: Evaluates as `True` only if the field of the given model is greater or equal (>=) to the
        parameter.
    * `EQUAL`: Evaluates as `True` only if the field of the given model is equal (==) to the parameter.
    * `NOT_EQUAL`: Evaluates as `True` only if the field of the given model is not equal (!=) to the parameter.
    * `IN`: Evaluates as `True` only if the field of the given model belongs (in) to the parameter (which must be a
        collection).
    * `LIKE`: Evaluates as `True` only if the field of the given model matches to the parameter _pattern.


    For example, to define a condition in which the `year` must be between `1994` and `2003` or the `color` must be
    `blue`, the condition can be writen as:

    .. code-block::

        Condition.OR(
            Condition.AND(Condition.GREATER_EQUAL("year", 1994), Condition.LOWER("year", 2003)),
            Condition.EQUAL("color", "blue")
        )

    """

    TRUE = _TRUE_CONDITION
    FALSE = _FALSE_CONDITION
    AND = _AndCondition
    OR = _OrCondition
    NOT = _NotCondition
    LOWER = _LowerCondition
    LOWER_EQUAL = _LowerEqualCondition
    GREATER = _GreaterCondition
    GREATER_EQUAL = _GreaterEqualCondition
    EQUAL = _EqualCondition
    NOT_EQUAL = _NotEqualCondition
    IN = _InCondition
    CONTAINS = _ContainsCondition
    LIKE = _LikeCondition


class _Ordering:
    def __init__(self, by: str, reverse: bool):
        self.by = by
        self.reverse = reverse

    def __eq__(self, other) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable[Any]:
        yield from (
            self.by,
            self.reverse,
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({', '.join(map(repr, self))})"


class Ordering:
    """Ordering class.

    This class provides the way to define ordering strategies for ``Model`` instances through the ``ASC`` and ``DESC``
    class methods, which retrieves instances containing the given information.

    For example, to define a descending ordering strategy by the `name` field:

    .. code-block::

        Ordering.DESC("name")

    """

    ASC = partial(_Ordering, reverse=False)
    DESC = partial(_Ordering, reverse=True)



from __future__ import (
    annotations,
)

from typing import (
    TYPE_CHECKING,
)

from minos.common import (
    MinosException,
)

if TYPE_CHECKING:
    from .deltas import (
        Delta,
    )
    from .entities import (
        Entity,
    )


class AggregateException(MinosException):
    """Base Aggregate module exception"""


class DeltaRepositoryException(AggregateException):
    """Base delta repository exception."""


class DeltaRepositoryConflictException(DeltaRepositoryException):
    """Exception to be raised when some ``DeltaEntry`` raises a conflict."""

    def __init__(self, error_message: str, offset: int):
        super().__init__(error_message)
        self.offset = offset


class SnapshotRepositoryException(AggregateException):
    """Base snapshot exception."""


class SnapshotRepositoryConflictException(SnapshotRepositoryException):
    """Exception to be raised when current version is newer than the one to be processed."""

    def __init__(self, previous: Entity, delta: Delta):
        self.previous = previous
        self.delta = delta
        super().__init__(
            f"Version for {repr(previous.classname)} root entity must be "
            f"greater than {previous.version}. Obtained: {delta.version}"
        )


class NotFoundException(SnapshotRepositoryException):
    """Exception to be raised when a ``Entity`` is not found on the repository."""


class AlreadyDeletedException(SnapshotRepositoryException):
    """Exception to be raised when a ``Entity`` is already deleted from the repository."""


class ValueObjectException(AggregateException):
    """If an attribute of an immutable class is modified, this exception will be raised"""


class RefException(AggregateException):
    """Exception to be raised when some reference can not be resolved."""



from __future__ import (
    annotations,
)

import logging
from datetime import (
    datetime,
)
from typing import (
    Type,
    TypeVar,
)
from uuid import (
    UUID,
)

from minos.common import (
    NULL_DATETIME,
    NULL_UUID,
    DeclarativeModel,
)

from ..deltas import (
    Delta,
    IncrementalFieldDiff,
)

logger = logging.getLogger(__name__)


class Entity(DeclarativeModel):
    """Entity class."""

    uuid: UUID

    version: int
    created_at: datetime
    updated_at: datetime

    def __init__(
        self,
        *args,
        uuid: UUID = NULL_UUID,
        version: int = 0,
        created_at: datetime = NULL_DATETIME,
        updated_at: datetime = NULL_DATETIME,
        **kwargs,
    ):

        super().__init__(uuid, version, created_at, updated_at, *args, **kwargs)

    def diff(self, another: Entity) -> Delta:
        """Compute the difference with another instance.

        Both ``Entity`` instances (``self`` and ``another``) must share the same ``uuid`` value.

        :param another: Another ``Entity`` instance.
        :return: An ``FieldDiffContainer`` instance.
        """
        return Delta.from_difference(self, another)

    def apply_diff(self, delta: Delta) -> None:
        """Apply the differences over the instance.

        :param delta: The ``FieldDiffContainer`` containing the values to be set.
        :return: This method does not return anything.
        """
        if self.uuid != delta.uuid:
            raise ValueError(
                f"To apply the difference, it must have same uuid. " f"Expected: {self.uuid!r} Obtained: {delta.uuid!r}"
            )

        logger.debug(f"Applying {delta!r} to {self!r}...")
        for diff in delta.fields_diff.flatten_values():
            if isinstance(diff, IncrementalFieldDiff):
                container = getattr(self, diff.name)
                if diff.action.is_delete:
                    container.discard(diff.value)
                else:
                    container.add(diff.value)
            else:
                setattr(self, diff.name, diff.value)
        self.version = delta.version
        self.updated_at = delta.created_at

    @classmethod
    def from_diff(cls: Type[T], delta: Delta, *args, **kwargs) -> T:
        """Build a new instance from an ``Delta``.

        :param delta: The difference that contains the data.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A new ``Entity`` instance.
        """
        return cls(
            *args,
            uuid=delta.uuid,
            version=delta.version,
            created_at=delta.created_at,
            updated_at=delta.created_at,
            **delta.get_fields(),
            **kwargs,
        )


T = TypeVar("T", bound=Entity)



from .collections import (
    EntitySet,
)
from .models import (
    Entity,
)
from .refs import (
    Ref,
    RefExtractor,
    RefInjector,
    RefResolver,
)
from .repositories import (
    EntityRepository,
)



from collections.abc import (
    Iterable,
)
from datetime import (
    datetime,
)
from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
)
from minos.common.testing import (
    MockedDatabaseClient,
    MockedDatabaseOperation,
)

from ....queries import (
    _Condition,
    _Ordering,
)
from ....snapshots import (
    SnapshotDatabaseOperationFactory,
)


class MockedSnapshotDatabaseOperationFactory(SnapshotDatabaseOperationFactory):
    """For testing purposes."""

    def build_create(self) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("create")

    def build_delete(self, transaction_uuids: Iterable[UUID]) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("delete")

    def build_submit(
        self,
        uuid: UUID,
        name: str,
        version: int,
        schema: bytes,
        data: dict[str, Any],
        created_at: datetime,
        updated_at: datetime,
        transaction_uuid: UUID,
    ) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("insert")

    def build_query(
        self,
        name: str,
        condition: _Condition,
        ordering: Optional[_Ordering],
        limit: Optional[int],
        transaction_uuids: tuple[UUID, ...],
        exclude_deleted: bool,
    ) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("query")

    def build_submit_offset(self, value: int) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("store_offset")

    def build_query_offset(self) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("get_offset")


MockedDatabaseClient.set_factory(SnapshotDatabaseOperationFactory, MockedSnapshotDatabaseOperationFactory)



from contextvars import (
    ContextVar,
)
from typing import (
    Final,
)

IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR: Final[ContextVar[bool]] = ContextVar(
    "is_repository_serialization", default=False
)
"""Context variable containing ``True`` if serialization has been started by a repository, or ``False`` otherwise."""



from __future__ import (
    annotations,
)

from collections import (
    defaultdict,
)
from itertools import (
    count,
)
from typing import (
    AsyncIterator,
    Optional,
)
from uuid import (
    UUID,
    uuid4,
)

from minos.common import (
    NULL_UUID,
    current_datetime,
)

from ...exceptions import (
    DeltaRepositoryConflictException,
)
from ..entries import (
    DeltaEntry,
)
from .abc import (
    DeltaRepository,
)


class InMemoryDeltaRepository(DeltaRepository):
    """Memory-based implementation of the delta repository class in ``minos``."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._storage = list()
        self._id_generator = count()
        self._next_versions = defaultdict(int)

    async def _submit(self, entry: DeltaEntry, **kwargs) -> DeltaEntry:
        if entry.uuid == NULL_UUID:
            entry.uuid = uuid4()

        next_version = self._get_next_version_id(entry)
        if entry.version is None:
            entry.version = next_version
        if entry.version < next_version:
            raise DeltaRepositoryConflictException(
                f"{entry!r} could not be submitted due to a key (uuid, version, transaction) collision",
                await self.offset,
            )

        if entry.created_at is None:
            entry.created_at = current_datetime()

        entry.id = self._generate_next_id()
        self._storage.append(entry)
        return entry

    def _generate_next_id(self) -> int:
        return next(self._id_generator) + 1

    def _get_next_version_id(self, entry: DeltaEntry) -> int:
        key = (entry.name, entry.uuid, entry.transaction_uuid)
        self._next_versions[key] += 1
        return self._next_versions[key]

    async def _select(
        self,
        uuid: Optional[int] = None,
        name: Optional[str] = None,
        version: Optional[int] = None,
        version_lt: Optional[int] = None,
        version_gt: Optional[int] = None,
        version_le: Optional[int] = None,
        version_ge: Optional[int] = None,
        id: Optional[int] = None,
        id_lt: Optional[int] = None,
        id_gt: Optional[int] = None,
        id_le: Optional[int] = None,
        id_ge: Optional[int] = None,
        transaction_uuid: Optional[UUID] = None,
        transaction_uuid_ne: Optional[UUID] = None,
        transaction_uuid_in: Optional[tuple[UUID, ...]] = None,
        *args,
        **kwargs,
    ) -> AsyncIterator[DeltaEntry]:

        # noinspection DuplicatedCode
        def _fn_filter(entry: DeltaEntry) -> bool:
            if uuid is not None and uuid != entry.uuid:
                return False
            if name is not None and name != entry.name:
                return False
            if version is not None and version != entry.version:
                return False
            if version_lt is not None and version_lt <= entry.version:
                return False
            if version_gt is not None and version_gt >= entry.version:
                return False
            if version_le is not None and version_le < entry.version:
                return False
            if version_ge is not None and version_ge > entry.version:
                return False
            if id is not None and id != entry.id:
                return False
            if id_lt is not None and id_lt <= entry.id:
                return False
            if id_gt is not None and id_gt >= entry.id:
                return False
            if id_le is not None and id_le < entry.id:
                return False
            if id_ge is not None and id_ge > entry.id:
                return False
            if transaction_uuid is not None and transaction_uuid != entry.transaction_uuid:
                return False
            if transaction_uuid_ne is not None and transaction_uuid_ne == entry.transaction_uuid:
                return False
            if transaction_uuid_in is not None and entry.transaction_uuid not in transaction_uuid_in:
                return False
            return True

        iterable = iter(self._storage)
        iterable = filter(_fn_filter, iterable)
        for item in iterable:
            yield item

    @property
    async def _offset(self) -> int:
        return len(self._storage)



from .entries import (
    SnapshotEntry,
)
from .repositories import (
    DatabaseSnapshotRepository,
    InMemorySnapshotRepository,
    SnapshotDatabaseOperationFactory,
    SnapshotRepository,
)
from .services import (
    SnapshotService,
)



from __future__ import (
    annotations,
)

from typing import (
    AsyncIterator,
    Optional,
)

from minos.common import (
    NULL_UUID,
    DatabaseMixin,
    DatabaseOperation,
    IntegrityException,
)

from ....exceptions import (
    DeltaRepositoryConflictException,
)
from ...entries import (
    DeltaEntry,
)
from ..abc import (
    DeltaRepository,
)
from .factories import (
    DeltaDatabaseOperationFactory,
)


class DatabaseDeltaRepository(DatabaseMixin[DeltaDatabaseOperationFactory], DeltaRepository):
    """Database-based implementation of the delta repository class."""

    def __init__(self, *args, database_key: Optional[tuple[str]] = None, **kwargs):
        if database_key is None:
            database_key = ("aggregate", "delta")
        super().__init__(*args, database_key=database_key, **kwargs)

    async def _setup(self) -> None:
        await super()._setup()
        operation = self.database_operation_factory.build_create()
        await self.execute_on_database(operation)

    async def _submit(self, entry: DeltaEntry, **kwargs) -> DeltaEntry:
        operation = await self._build_submit_operation(entry)

        try:
            response = await self.execute_on_database_and_fetch_one(operation)
        except IntegrityException:
            raise DeltaRepositoryConflictException(
                f"{entry!r} could not be submitted due to a key (uuid, version, transaction) collision",
                await self.offset,
            )

        entry.id, entry.uuid, entry.version, entry.created_at = response
        return entry

    async def _build_submit_operation(self, entry: DeltaEntry) -> DatabaseOperation:
        lock = None
        if entry.uuid != NULL_UUID:
            lock = entry.uuid.int & (1 << 32) - 1

        if entry.transaction_uuid != NULL_UUID:
            transaction = await self.transaction_repository.get(uuid=entry.transaction_uuid)
            transaction_uuids = await transaction.uuids
        else:
            transaction_uuids = (NULL_UUID,)

        return self.database_operation_factory.build_submit(
            transaction_uuids=transaction_uuids, **entry.as_raw(), lock=lock
        )

    async def _select(self, streaming_mode: Optional[bool] = None, **kwargs) -> AsyncIterator[DeltaEntry]:
        operation = self.database_operation_factory.build_query(**kwargs)
        async for row in self.execute_on_database_and_fetch_all(operation, streaming_mode=streaming_mode):
            yield DeltaEntry(*row)

    @property
    async def _offset(self) -> int:
        operation = self.database_operation_factory.build_query_offset()
        row = await self.execute_on_database_and_fetch_one(operation)
        return row[0] or 0



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    gather,
)
from typing import (
    TYPE_CHECKING,
)
from uuid import (
    UUID,
)

from cached_property import (
    cached_property,
)

from minos.common import (
    Config,
    Inject,
    ModelType,
)
from minos.networks import (
    EnrouteDecorator,
    Request,
    Response,
    ResponseException,
    enroute,
)

from .repositories import (
    SnapshotRepository,
)

if TYPE_CHECKING:
    from ..entities import (
        Entity,
    )

logger = logging.getLogger(__name__)


class SnapshotService:
    """Snapshot Service class."""

    # noinspection PyUnusedLocal
    @Inject()
    def __init__(
        self,
        *args,
        config: Config,
        snapshot_repository: SnapshotRepository,
        **kwargs,
    ):
        self.config = config
        self.snapshot_repository = snapshot_repository

    @classmethod
    def __get_enroute__(cls, config: Config) -> dict[str, set[EnrouteDecorator]]:
        from ..entities import (
            RefResolver,
        )

        aggregate_config = config.get_aggregate()
        root_entity = aggregate_config["entities"][0]
        name = root_entity.__name__

        return {
            cls.__get_many__.__name__: {enroute.broker.command(RefResolver.build_topic_name(name))},
            cls.__synchronize__.__name__: {enroute.periodic.event("* * * * *")},
        }

    async def __get_many__(self, request: Request) -> Response:
        """Get many ``Entity`` instances.

        :param request: The ``Request`` instance that contains the instance identifiers.
        :return: A ``Response`` instance containing the requested instances.
        """
        try:
            content = await request.content(model_type=ModelType.build("Query", {"uuids": list[UUID]}))
        except Exception as exc:
            raise ResponseException(f"There was a problem while parsing the given request: {exc!r}")

        try:
            futures = (self.snapshot_repository.get(self.type_, uuid) for uuid in content["uuids"])
            instances = await gather(*futures)
        except Exception as exc:
            raise ResponseException(f"There was a problem while getting the instances: {exc!r}")

        return Response(instances)

    @cached_property
    def type_(self) -> type[Entity]:
        """Load the concrete ``Entity`` class.

        :return: A ``Type`` object.
        """
        aggregate_config = self.config.get_aggregate()
        return aggregate_config["entities"][0]

    # noinspection PyUnusedLocal
    async def __synchronize__(self, request: Request) -> None:
        """Performs a Snapshot synchronization every minute.

        :param request: A request containing information related with scheduling.
        :return: This method does not return anything.
        """
        logger.info("Performing periodic Snapshot synchronization...")
        await self.snapshot_repository.synchronize()



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from datetime import (
    timedelta,
)
from typing import (
    Optional,
)
from unittest.mock import (
    MagicMock,
    call,
)
from uuid import (
    uuid4,
)

from minos.aggregate import (
    Action,
    AlreadyDeletedException,
    Condition,
    DeltaEntry,
    Entity,
    FieldDiff,
    FieldDiffContainer,
    NotFoundException,
    Ordering,
    Ref,
    SnapshotEntry,
    SnapshotRepository,
)
from minos.common import (
    classname,
    current_datetime,
)
from minos.common.testing import (
    MinosTestCase,
)
from minos.transactions import (
    TransactionEntry,
    TransactionStatus,
)


class SnapshotRepositoryTestCase(MinosTestCase, ABC):
    __test__ = False

    snapshot_repository: SnapshotRepository

    class NumbersList(Entity):
        """For testing purposes"""

        numbers: list[int]

    class Number(Entity):
        """For testing purposes"""

        value: int

    class Owner(Entity):
        """For testing purposes"""

        name: str
        surname: str
        age: Optional[int]

    class Car(Entity):
        """For testing purposes"""

        doors: int
        color: str
        owner: Optional[Ref[SnapshotRepositoryTestCase.Owner]]

    def setUp(self) -> None:
        super().setUp()
        self.snapshot_repository = self.build_snapshot_repository()

        self.uuid_1 = uuid4()
        self.uuid_2 = uuid4()
        self.uuid_3 = uuid4()

        self.transaction_1 = uuid4()
        self.transaction_2 = uuid4()
        self.transaction_3 = uuid4()
        self.transaction_4 = uuid4()

    @abstractmethod
    def build_snapshot_repository(self) -> SnapshotRepository:
        pass

    async def populate(self) -> None:
        diff = FieldDiffContainer([FieldDiff("doors", int, 3), FieldDiff("color", str, "blue")])
        name: str = classname(self.Car)

        await self.delta_repository.create(DeltaEntry(self.uuid_1, name, 1, diff.avro_bytes))
        await self.delta_repository.update(DeltaEntry(self.uuid_1, name, 2, diff.avro_bytes))
        await self.delta_repository.create(DeltaEntry(self.uuid_2, name, 1, diff.avro_bytes))
        await self.delta_repository.update(DeltaEntry(self.uuid_1, name, 3, diff.avro_bytes))
        await self.delta_repository.delete(DeltaEntry(self.uuid_1, name, 4))
        await self.delta_repository.update(DeltaEntry(self.uuid_2, name, 2, diff.avro_bytes))
        await self.delta_repository.update(
            DeltaEntry(self.uuid_2, name, 3, diff.avro_bytes, transaction_uuid=self.transaction_1)
        )
        await self.delta_repository.delete(
            DeltaEntry(self.uuid_2, name, 3, bytes(), transaction_uuid=self.transaction_2)
        )
        await self.delta_repository.update(
            DeltaEntry(self.uuid_2, name, 4, diff.avro_bytes, transaction_uuid=self.transaction_1)
        )
        await self.delta_repository.create(DeltaEntry(self.uuid_3, name, 1, diff.avro_bytes))
        await self.delta_repository.delete(
            DeltaEntry(self.uuid_2, name, 3, bytes(), transaction_uuid=self.transaction_3)
        )
        await self.transaction_repository.submit(TransactionEntry(self.transaction_1, TransactionStatus.PENDING))
        await self.transaction_repository.submit(TransactionEntry(self.transaction_2, TransactionStatus.PENDING))
        await self.transaction_repository.submit(TransactionEntry(self.transaction_3, TransactionStatus.REJECTED))
        await self.transaction_repository.submit(
            TransactionEntry(self.transaction_4, TransactionStatus.REJECTED, self.transaction_3)
        )

    async def populate_and_synchronize(self):
        await self.populate()
        await self.synchronize()

    async def synchronize(self):
        await self.snapshot_repository.synchronize()

    async def asyncSetUp(self):
        await super().asyncSetUp()
        await self.snapshot_repository.setup()

    async def asyncTearDown(self):
        await self.snapshot_repository.destroy()
        await super().asyncTearDown()

    def assert_equal_snapshot_entries(self, expected: list[SnapshotEntry], observed: list[SnapshotEntry]):
        self.assertEqual(len(expected), len(observed))
        for exp, obs in zip(expected, observed):
            if exp.data is None:
                with self.assertRaises(AlreadyDeletedException):
                    # noinspection PyStatementEffect
                    obs.build()
            else:
                self.assertEqual(exp.build(), obs.build())
            self.assertAlmostEqual(exp.created_at or current_datetime(), obs.created_at, delta=timedelta(seconds=5))
            self.assertAlmostEqual(exp.updated_at or current_datetime(), obs.updated_at, delta=timedelta(seconds=5))

    def test_type(self):
        self.assertTrue(isinstance(self.snapshot_repository, SnapshotRepository))

    async def test_dispatch(self):
        await self.populate_and_synchronize()

        # noinspection PyTypeChecker
        iterable = self.snapshot_repository.find_entries(
            self.Car.classname, Condition.TRUE, Ordering.ASC("updated_at"), exclude_deleted=False
        )
        observed = [v async for v in iterable]

        # noinspection PyTypeChecker
        expected = [
            SnapshotEntry(self.uuid_1, self.Car.classname, 4),
            SnapshotEntry.from_entity(
                self.Car(
                    3,
                    "blue",
                    uuid=self.uuid_2,
                    version=2,
                    created_at=observed[1].created_at,
                    updated_at=observed[1].updated_at,
                )
            ),
            SnapshotEntry.from_entity(
                self.Car(
                    3,
                    "blue",
                    uuid=self.uuid_3,
                    version=1,
                    created_at=observed[2].created_at,
                    updated_at=observed[2].updated_at,
                )
            ),
        ]
        self.assert_equal_snapshot_entries(expected, observed)

    async def test_dispatch_first_transaction(self):
        await self.populate_and_synchronize()

        # noinspection PyTypeChecker
        iterable = self.snapshot_repository.find_entries(
            self.Car.classname,
            Condition.TRUE,
            Ordering.ASC("updated_at"),
            exclude_deleted=False,
            transaction=TransactionEntry(self.transaction_1),
        )
        observed = [v async for v in iterable]

        # noinspection PyTypeChecker
        expected = [
            SnapshotEntry(self.uuid_1, self.Car.classname, 4),
            SnapshotEntry.from_entity(
                self.Car(
                    3,
                    "blue",
                    uuid=self.uuid_2,
                    version=4,
                    created_at=observed[1].created_at,
                    updated_at=observed[1].updated_at,
                )
            ),
            SnapshotEntry.from_entity(
                self.Car(
                    3,
                    "blue",
                    uuid=self.uuid_3,
                    version=1,
                    created_at=observed[2].created_at,
                    updated_at=observed[2].updated_at,
                )
            ),
        ]
        self.assert_equal_snapshot_entries(expected, observed)

    async def test_dispatch_second_transaction(self):
        await self.populate_and_synchronize()

        # noinspection PyTypeChecker
        iterable = self.snapshot_repository.find_entries(
            self.Car.classname,
            Condition.TRUE,
            Ordering.ASC("updated_at"),
            exclude_deleted=False,
            transaction=TransactionEntry(self.transaction_2),
        )
        observed = [v async for v in iterable]

        # noinspection PyTypeChecker
        expected = [
            SnapshotEntry(self.uuid_1, self.Car.classname, 4),
            SnapshotEntry(self.uuid_2, self.Car.classname, 4),
            SnapshotEntry.from_entity(
                self.Car(
                    3,
                    "blue",
                    uuid=self.uuid_3,
                    version=1,
                    created_at=observed[2].created_at,
                    updated_at=observed[2].updated_at,
                )
            ),
        ]
        self.assert_equal_snapshot_entries(expected, observed)

    async def test_dispatch_third_transaction(self):
        await self.populate_and_synchronize()

        # noinspection PyTypeChecker
        iterable = self.snapshot_repository.find_entries(
            self.Car.classname,
            Condition.TRUE,
            Ordering.ASC("updated_at"),
            exclude_deleted=False,
            transaction_uuid=self.transaction_3,
        )
        observed = [v async for v in iterable]

        # noinspection PyTypeChecker
        expected = [
            SnapshotEntry(self.uuid_1, self.Car.classname, 4),
            SnapshotEntry.from_entity(
                self.Car(
                    3,
                    "blue",
                    uuid=self.uuid_2,
                    version=2,
                    created_at=observed[1].created_at,
                    updated_at=observed[1].updated_at,
                )
            ),
            SnapshotEntry.from_entity(
                self.Car(
                    3,
                    "blue",
                    uuid=self.uuid_3,
                    version=1,
                    created_at=observed[2].created_at,
                    updated_at=observed[2].updated_at,
                )
            ),
        ]
        self.assert_equal_snapshot_entries(expected, observed)

    async def test_dispatch_ignore_previous_version(self):
        await self.populate()
        diff = FieldDiffContainer([FieldDiff("doors", int, 3), FieldDiff("color", str, "blue")])
        # noinspection PyTypeChecker
        name: str = self.Car.classname
        condition = Condition.EQUAL("uuid", self.uuid_1)

        async def _fn(*args, id_gt: Optional[int] = None, **kwargs):
            if id_gt is not None and id_gt > 0:
                return
            yield DeltaEntry(self.uuid_1, name, 1, diff.avro_bytes, 1, Action.CREATE, current_datetime())
            yield DeltaEntry(self.uuid_1, name, 3, diff.avro_bytes, 2, Action.CREATE, current_datetime())
            yield DeltaEntry(self.uuid_1, name, 2, diff.avro_bytes, 3, Action.CREATE, current_datetime())

        self.delta_repository.select = MagicMock(side_effect=_fn)
        await self.snapshot_repository.synchronize()

        observed = [v async for v in self.snapshot_repository.find_entries(name, condition)]

        # noinspection PyTypeChecker
        expected = [
            SnapshotEntry(
                uuid=self.uuid_1,
                name=name,
                version=3,
                schema=self.Car.avro_schema,
                data=self.Car(3, "blue", uuid=self.uuid_1, version=1).avro_data,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            )
        ]
        self.assert_equal_snapshot_entries(expected, observed)

    async def test_dispatch_with_offset(self):
        await self.populate()

        mock = MagicMock(side_effect=self.delta_repository.select)
        self.delta_repository.select = mock

        await self.snapshot_repository.synchronize()
        self.assertEqual(1, mock.call_count)
        self.assertEqual(call(id_gt=0, synchronize=False), mock.call_args)
        mock.reset_mock()

        # noinspection PyTypeChecker
        entry = DeltaEntry(
            uuid=self.uuid_3,
            name=self.Car.classname,
            data=FieldDiffContainer([FieldDiff("doors", int, 3), FieldDiff("color", str, "blue")]).avro_bytes,
        )
        await self.delta_repository.create(entry)

        await self.snapshot_repository.synchronize()
        self.assertEqual(1, mock.call_count)
        self.assertEqual(call(id_gt=11, synchronize=False), mock.call_args)
        mock.reset_mock()

        await self.snapshot_repository.synchronize()
        self.assertEqual(1, mock.call_count)
        self.assertEqual(call(id_gt=12, synchronize=False), mock.call_args)
        mock.reset_mock()

        await self.snapshot_repository.synchronize()
        self.assertEqual(1, mock.call_count)
        self.assertEqual(call(id_gt=12, synchronize=False), mock.call_args)
        mock.reset_mock()

    async def test_find_by_uuid(self):
        await self.populate_and_synchronize()
        condition = Condition.IN("uuid", [self.uuid_2, self.uuid_3])

        iterable = self.snapshot_repository.find(self.Car, condition, ordering=Ordering.ASC("updated_at"))
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_2,
                version=2,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[1].created_at,
                updated_at=observed[1].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)

    async def test_find_contains(self):
        a = FieldDiffContainer([FieldDiff("numbers", list[int], [1, 2, 3])])
        b = FieldDiffContainer([FieldDiff("numbers", list[int], [4, 5, 6])])
        c = FieldDiffContainer([FieldDiff("numbers", list[int], [3, 8, 9])])
        await self.delta_repository.create(
            DeltaEntry(name=self.NumbersList.classname, data=a.avro_bytes, uuid=self.uuid_1)
        )
        await self.delta_repository.create(
            DeltaEntry(name=self.NumbersList.classname, data=b.avro_bytes, uuid=self.uuid_2)
        )
        await self.delta_repository.create(
            DeltaEntry(name=self.NumbersList.classname, data=c.avro_bytes, uuid=self.uuid_3)
        )
        await self.synchronize()

        condition = Condition.CONTAINS("numbers", 3)

        iterable = self.snapshot_repository.find(self.NumbersList, condition, ordering=Ordering.ASC("updated_at"))
        observed = [v async for v in iterable]

        expected = [
            await self.snapshot_repository.get(self.NumbersList, self.uuid_1),
            await self.snapshot_repository.get(self.NumbersList, self.uuid_3),
        ]
        self.assertEqual(expected, observed)

    async def test_find_equal(self):
        a = FieldDiffContainer([FieldDiff("value", int, 1)])
        b = FieldDiffContainer([FieldDiff("value", int, 2)])
        c = FieldDiffContainer([FieldDiff("value", int, 1)])
        await self.delta_repository.create(DeltaEntry(name=self.Number.classname, data=a.avro_bytes, uuid=self.uuid_1))
        await self.delta_repository.create(DeltaEntry(name=self.Number.classname, data=b.avro_bytes, uuid=self.uuid_2))
        await self.delta_repository.create(DeltaEntry(name=self.Number.classname, data=c.avro_bytes, uuid=self.uuid_3))
        await self.synchronize()

        condition = Condition.EQUAL("value", 1)

        iterable = self.snapshot_repository.find(self.Number, condition, ordering=Ordering.ASC("updated_at"))
        observed = [v async for v in iterable]

        expected = [
            await self.snapshot_repository.get(self.Number, self.uuid_1),
            await self.snapshot_repository.get(self.Number, self.uuid_3),
        ]
        self.assertEqual(expected, observed)

    async def test_find_with_transaction(self):
        await self.populate_and_synchronize()
        condition = Condition.IN("uuid", [self.uuid_2, self.uuid_3])

        iterable = self.snapshot_repository.find(
            self.Car,
            condition,
            ordering=Ordering.ASC("updated_at"),
            transaction=TransactionEntry(self.transaction_1),
        )
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_2,
                version=4,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[1].created_at,
                updated_at=observed[1].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)

    async def test_find_with_transaction_delete(self):
        await self.populate_and_synchronize()
        condition = Condition.IN("uuid", [self.uuid_2, self.uuid_3])

        iterable = self.snapshot_repository.find(
            self.Car,
            condition,
            ordering=Ordering.ASC("updated_at"),
            transaction=TransactionEntry(self.transaction_2),
        )
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)

    async def test_find_with_transaction_reverted(self):
        await self.populate_and_synchronize()
        condition = Condition.IN("uuid", [self.uuid_2, self.uuid_3])

        iterable = self.snapshot_repository.find(
            self.Car,
            condition,
            ordering=Ordering.ASC("updated_at"),
            transaction=TransactionEntry(self.transaction_4),
        )
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_2,
                version=2,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[1].created_at,
                updated_at=observed[1].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)

    async def test_find_streaming_true(self):
        await self.populate_and_synchronize()
        condition = Condition.IN("uuid", [self.uuid_2, self.uuid_3])

        iterable = self.snapshot_repository.find(
            self.Car, condition, streaming_mode=True, ordering=Ordering.ASC("updated_at")
        )
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_2,
                version=2,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[1].created_at,
                updated_at=observed[1].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)

    async def test_find_with_duplicates(self):
        await self.populate_and_synchronize()
        uuids = [self.uuid_2, self.uuid_2, self.uuid_3]
        condition = Condition.IN("uuid", uuids)

        iterable = self.snapshot_repository.find(self.Car, condition, ordering=Ordering.ASC("updated_at"))
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_2,
                version=2,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[1].created_at,
                updated_at=observed[1].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)

    async def test_find_empty(self):
        await self.populate_and_synchronize()
        observed = {v async for v in self.snapshot_repository.find(self.Car, Condition.FALSE)}

        expected = set()
        self.assertEqual(expected, observed)

    async def test_find_one(self):
        await self.populate_and_synchronize()
        condition = Condition.IN("uuid", {self.uuid_2})

        observed = await self.snapshot_repository.find_one(self.Car, condition)

        expected = self.Car(
            3,
            "blue",
            uuid=self.uuid_2,
            version=2,
            created_at=observed.created_at,
            updated_at=observed.updated_at,
        )
        self.assertEqual(expected, observed)

    async def test_find_one_raises(self):
        await self.populate_and_synchronize()
        condition = Condition.FALSE

        with self.assertRaises(NotFoundException):
            await self.snapshot_repository.find_one(self.Car, condition)

    async def test_get_all(self):
        await self.populate_and_synchronize()
        iterable = self.snapshot_repository.find(self.Car, Condition.TRUE, ordering=Ordering.ASC("updated_at"))
        expected = [v async for v in iterable]

        iterable = self.snapshot_repository.get_all(self.Car, ordering=Ordering.ASC("updated_at"))
        observed = [v async for v in iterable]

        self.assertEqual(expected, observed)

    async def test_get(self):
        await self.populate_and_synchronize()
        observed = await self.snapshot_repository.get(self.Car, self.uuid_2)

        expected = self.Car(
            3, "blue", uuid=self.uuid_2, version=2, created_at=observed.created_at, updated_at=observed.updated_at
        )
        self.assertEqual(expected, observed)

    async def test_get_with_transaction(self):
        await self.populate_and_synchronize()

        observed = await self.snapshot_repository.get(
            self.Car, self.uuid_2, transaction=TransactionEntry(self.transaction_1)
        )

        expected = self.Car(
            3, "blue", uuid=self.uuid_2, version=4, created_at=observed.created_at, updated_at=observed.updated_at
        )
        self.assertEqual(expected, observed)

    async def test_get_raises(self):
        await self.populate_and_synchronize()
        with self.assertRaises(AlreadyDeletedException):
            await self.snapshot_repository.get(self.Car, self.uuid_1)
        with self.assertRaises(NotFoundException):
            await self.snapshot_repository.get(self.Car, uuid4())

    async def test_get_with_transaction_raises(self):
        await self.populate_and_synchronize()
        with self.assertRaises(AlreadyDeletedException):
            await self.snapshot_repository.get(self.Car, self.uuid_2, transaction=TransactionEntry(self.transaction_2))

    async def test_find(self):
        await self.populate_and_synchronize()
        condition = Condition.EQUAL("color", "blue")
        iterable = self.snapshot_repository.find(self.Car, condition, ordering=Ordering.ASC("updated_at"))
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_2,
                version=2,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[1].created_at,
                updated_at=observed[1].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)

    async def test_find_all(self):
        await self.populate_and_synchronize()
        iterable = self.snapshot_repository.find(self.Car, Condition.TRUE, Ordering.ASC("updated_at"))
        observed = [v async for v in iterable]

        expected = [
            self.Car(
                3,
                "blue",
                uuid=self.uuid_2,
                version=2,
                created_at=observed[0].created_at,
                updated_at=observed[0].updated_at,
            ),
            self.Car(
                3,
                "blue",
                uuid=self.uuid_3,
                version=1,
                created_at=observed[1].created_at,
                updated_at=observed[1].updated_at,
            ),
        ]
        self.assertEqual(expected, observed)



"""The Aggregate pattern of the Minos Framework."""

__author__ = "Minos Framework Devs"
__email__ = "hey@minos.run"
__version__ = "0.8.0.dev2"

from .actions import (
    Action,
)
from .collections import (
    IncrementalSet,
    IncrementalSetDiff,
    IncrementalSetDiffEntry,
)
from .contextvars import (
    IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR,
)
from .deltas import (
    DatabaseDeltaRepository,
    Delta,
    DeltaDatabaseOperationFactory,
    DeltaEntry,
    DeltaRepository,
    FieldDiff,
    FieldDiffContainer,
    IncrementalFieldDiff,
    InMemoryDeltaRepository,
)
from .entities import (
    Entity,
    EntityRepository,
    EntitySet,
    Ref,
    RefExtractor,
    RefInjector,
    RefResolver,
)
from .exceptions import (
    AggregateException,
    AlreadyDeletedException,
    DeltaRepositoryConflictException,
    DeltaRepositoryException,
    NotFoundException,
    RefException,
    SnapshotRepositoryConflictException,
    SnapshotRepositoryException,
    ValueObjectException,
)
from .impl import (
    Aggregate,
)
from .queries import (
    Condition,
    Ordering,
)
from .snapshots import (
    DatabaseSnapshotRepository,
    InMemorySnapshotRepository,
    SnapshotDatabaseOperationFactory,
    SnapshotEntry,
    SnapshotRepository,
    SnapshotService,
)
from .value_objects import (
    ValueObject,
    ValueObjectSet,
)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from contextlib import (
    suppress,
)
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    Inject,
    Injectable,
    Lock,
    LockPool,
    NotProvidedException,
    PoolFactory,
    classname,
)
from minos.transactions import (
    TRANSACTION_CONTEXT_VAR,
    TransactionalMixin,
    TransactionEntry,
    TransactionStatus,
)

from ...actions import (
    Action,
)
from ...contextvars import (
    IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR,
)
from ...exceptions import (
    DeltaRepositoryConflictException,
    DeltaRepositoryException,
)
from ..entries import (
    DeltaEntry,
)
from ..models import (
    Delta,
)

if TYPE_CHECKING:
    from ...entities import (
        Entity,
    )


@Injectable("delta_repository")
class DeltaRepository(ABC, TransactionalMixin):
    """Base delta repository class in ``minos``."""

    _lock_pool: LockPool

    @Inject()
    def __init__(
        self,
        lock_pool: Optional[LockPool] = None,
        pool_factory: Optional[PoolFactory] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if lock_pool is None and pool_factory is not None:
            lock_pool = pool_factory.get_pool("lock")

        if lock_pool is None:
            raise NotProvidedException("A lock pool instance is required.")

        self._lock_pool = lock_pool

    def transaction(self, **kwargs) -> TransactionEntry:
        """Build a transaction instance related to the repository.

        :param kwargs: Additional named arguments.
        :return: A new ``TransactionEntry`` instance.
        """
        return TransactionEntry(repository=self.transaction_repository, **kwargs)

    async def create(self, entry: Union[Delta, DeltaEntry]) -> DeltaEntry:
        """Store new creation entry into the repository.

        :param entry: Entry to be stored.
        :return: The repository entry containing the stored information.
        """

        entry.action = Action.CREATE
        return await self.submit(entry)

    async def update(self, entry: Union[Delta, DeltaEntry]) -> DeltaEntry:
        """Store new update entry into the repository.

        :param entry: Entry to be stored.
        :return: The repository entry containing the stored information.
        """

        entry.action = Action.UPDATE
        return await self.submit(entry)

    async def delete(self, entry: Union[Delta, DeltaEntry]) -> DeltaEntry:
        """Store new deletion entry into the repository.

        :param entry: Entry to be stored.
        :return: The repository entry containing the stored information.
        """

        entry.action = Action.DELETE
        return await self.submit(entry)

    async def submit(self, entry: Union[Delta, DeltaEntry], **kwargs) -> DeltaEntry:
        """Store new entry into the repository.

        :param entry: The entry to be stored.
        :param kwargs: Additional named arguments.
        :return: The repository entry containing the stored information.
        """

        token = IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR.set(True)
        try:
            transaction = TRANSACTION_CONTEXT_VAR.get()

            if isinstance(entry, Delta):
                entry = DeltaEntry.from_delta(entry, transaction=transaction)

            if not isinstance(entry.action, Action):
                raise DeltaRepositoryException("The 'DeltaEntry.action' attribute must be an 'Action' instance.")

            async with self.write_lock():
                if not await self.validate(entry, **kwargs):
                    raise DeltaRepositoryConflictException(f"{entry!r} could not be committed!", await self.offset)

                entry = await self._submit(entry, **kwargs)

        finally:
            IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR.reset(token)

        return entry

    # noinspection PyUnusedLocal
    async def validate(self, entry: DeltaEntry, transaction_uuid_ne: Optional[UUID] = None, **kwargs) -> bool:
        """Check if it is able to submit the given entry.

        :param entry: The entry to be validated.
        :param transaction_uuid_ne: Optional transaction identifier to skip it from the validation.
        :param kwargs: Additional named arguments.
        :return: ``True`` if the entry can be submitted or ``False`` otherwise.
        """
        iterable = self.transaction_repository.select(
            destination_uuid=entry.transaction_uuid,
            uuid_ne=transaction_uuid_ne,
            status_in=(TransactionStatus.RESERVING, TransactionStatus.RESERVED, TransactionStatus.COMMITTING),
        )

        transaction_uuids = {e.uuid async for e in iterable}

        if len(transaction_uuids):
            with suppress(StopAsyncIteration):
                iterable = self.select(uuid=entry.uuid, transaction_uuid_in=tuple(transaction_uuids), **kwargs)

                await iterable.__anext__()  # Will raise a `StopAsyncIteration` exception if not any item.

                return False

        return True

    @abstractmethod
    async def _submit(self, entry: DeltaEntry, **kwargs) -> DeltaEntry:
        raise NotImplementedError

    # noinspection PyShadowingBuiltins
    async def select(
        self,
        uuid: Optional[UUID] = None,
        name: Optional[Union[str, type[Entity]]] = None,
        version: Optional[int] = None,
        version_lt: Optional[int] = None,
        version_gt: Optional[int] = None,
        version_le: Optional[int] = None,
        version_ge: Optional[int] = None,
        id: Optional[int] = None,
        id_lt: Optional[int] = None,
        id_gt: Optional[int] = None,
        id_le: Optional[int] = None,
        id_ge: Optional[int] = None,
        transaction_uuid: Optional[UUID] = None,
        transaction_uuid_ne: Optional[UUID] = None,
        transaction_uuid_in: Optional[tuple[UUID, ...]] = None,
        **kwargs,
    ) -> AsyncIterator[DeltaEntry]:
        """Perform a selection query of entries stored in to the repository.

        :param uuid: The identifier must be equal to the given value.
        :param name: The classname must be equal to the given value.
        :param version: The version must be equal to the given value.
        :param version_lt: The version must be lower than the given value.
        :param version_gt: The version must be greater than the given value.
        :param version_le: The version must be lower or equal to the given value.
        :param version_ge: The version must be greater or equal to the given value.
        :param id: The entry identifier must be equal to the given value.
        :param id_lt: The entry identifier must be lower than the given value.
        :param id_gt: The entry identifier must be greater than the given value.
        :param id_le: The entry identifier must be lower or equal to the given value.
        :param id_ge: The entry identifier must be greater or equal to the given value.
        :param transaction_uuid: The transaction identifier must be equal to the given value.
        :param transaction_uuid_ne: The transaction identifier must be distinct of the given value.
        :param transaction_uuid_in: The destination transaction identifier must be equal to one of the given values.
        :return: A list of entries.
        """
        if isinstance(name, type):
            name = classname(name)
        generator = self._select(
            uuid=uuid,
            name=name,
            version=version,
            version_lt=version_lt,
            version_gt=version_gt,
            version_le=version_le,
            version_ge=version_ge,
            id=id,
            id_lt=id_lt,
            id_gt=id_gt,
            id_le=id_le,
            id_ge=id_ge,
            transaction_uuid=transaction_uuid,
            transaction_uuid_ne=transaction_uuid_ne,
            transaction_uuid_in=transaction_uuid_in,
            **kwargs,
        )
        # noinspection PyTypeChecker
        async for entry in generator:
            yield entry

    @abstractmethod
    async def _select(self, *args, **kwargs) -> AsyncIterator[DeltaEntry]:
        """Perform a selection query of entries stored in to the repository."""

    @property
    def offset(self) -> Awaitable[int]:
        """Get the current repository offset.

        :return: An awaitable containing an integer value.
        """
        return self._offset

    @property
    @abstractmethod
    async def _offset(self) -> int:
        raise NotImplementedError

    def write_lock(self) -> Lock:
        """Get a lock.

        :return: An asynchronous context manager.
        """
        return self._lock_pool.acquire("aggregate_delta_write_lock")

    async def get_collided_transactions(self, transaction_uuid: UUID) -> set[UUID]:
        """Get the set of collided transaction identifiers.

        :param transaction_uuid: The identifier of the transaction to be committed.
        :return: A ``set`` or ``UUID`` values.
        """
        entries = dict()
        async for entry in self.select(transaction_uuid=transaction_uuid):
            if entry.uuid in entries and entry.version < entries[entry.uuid]:
                continue
            entries[entry.uuid] = entry.version

        transaction_uuids = set()
        for uuid, version in entries.items():
            async for entry in self.select(uuid=uuid, version=version):
                if entry.transaction_uuid != transaction_uuid:
                    transaction_uuids.add(entry.transaction_uuid)
        return transaction_uuids

    async def commit_transaction(self, transaction_uuid: UUID, destination_transaction_uuid: UUID) -> None:
        """Commit the transaction with given identifier.

        :param transaction_uuid: The identifier of the transaction to be committed.
        :param destination_transaction_uuid: The identifier of the destination transaction.
        :return: This method does not return anything.
        """
        async for entry in self.select(transaction_uuid=transaction_uuid):
            new = DeltaEntry.from_another(entry, transaction_uuid=destination_transaction_uuid)
            await self.submit(new, transaction_uuid_ne=transaction_uuid)



from __future__ import (
    annotations,
)

from contextlib import (
    suppress,
)
from functools import (
    cmp_to_key,
)
from operator import (
    attrgetter,
)
from typing import (
    AsyncIterator,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    NULL_UUID,
    Inject,
    NotProvidedException,
)
from minos.transactions import (
    TransactionEntry,
    TransactionRepository,
    TransactionStatus,
)

from ...deltas import (
    DeltaEntry,
    DeltaRepository,
)
from ...exceptions import (
    AlreadyDeletedException,
)
from ...queries import (
    _Condition,
    _Ordering,
)
from ..entries import (
    SnapshotEntry,
)
from .abc import (
    SnapshotRepository,
)


class InMemorySnapshotRepository(SnapshotRepository):
    """InMemory Snapshot class.

    The snapshot provides a direct accessor to the ``Entity`` instances stored as deltas by the delta repository
    class.
    """

    @Inject()
    def __init__(
        self,
        *args,
        delta_repository: DeltaRepository,
        transaction_repository: TransactionRepository,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if delta_repository is None:
            raise NotProvidedException("An delta repository instance is required.")

        if transaction_repository is None:
            raise NotProvidedException("A transaction repository instance is required.")

        self._delta_repository = delta_repository
        self._transaction_repository = transaction_repository

    async def _find_entries(
        self,
        name: str,
        condition: _Condition,
        ordering: Optional[_Ordering],
        limit: Optional[int],
        exclude_deleted: bool,
        **kwargs,
    ) -> AsyncIterator[SnapshotEntry]:
        uuids = {v.uuid async for v in self._delta_repository.select(name=name)}

        entries = list()
        for uuid in uuids:
            entry = await self._get(name, uuid, **kwargs)

            try:
                instance = entry.build()
                if condition.evaluate(instance):
                    entries.append(entry)
            except AlreadyDeletedException:
                # noinspection PyTypeChecker
                if not exclude_deleted and condition.evaluate(entry):
                    entries.append(entry)

        if ordering is not None:

            def _cmp(a: SnapshotEntry, b: SnapshotEntry) -> int:
                with suppress(AlreadyDeletedException):
                    with suppress(AlreadyDeletedException):
                        try:
                            aa = attrgetter(ordering.by)(a.build())
                        except AlreadyDeletedException:
                            aa = attrgetter(ordering.by)(a)
                    with suppress(AlreadyDeletedException):
                        try:
                            bb = attrgetter(ordering.by)(b.build())
                        except AlreadyDeletedException:
                            bb = attrgetter(ordering.by)(b)

                    if aa > bb:
                        return 1
                    elif aa < bb:
                        return -1

                return 0

            entries.sort(key=cmp_to_key(_cmp), reverse=ordering.reverse)

        if limit is not None:
            entries = entries[:limit]

        for entry in entries:
            yield entry

    # noinspection PyMethodOverriding
    async def _get(
        self, name: str, uuid: UUID, transaction: Optional[TransactionEntry] = None, **kwargs
    ) -> SnapshotEntry:
        transaction_uuids = await self._get_transaction_uuids(transaction)
        entries = await self._get_delta_entries(name, uuid, transaction_uuids)
        return self._build_instance(entries, **kwargs)

    async def _get_transaction_uuids(self, transaction: Optional[TransactionEntry]) -> tuple[UUID, ...]:
        if transaction is None:
            transaction_uuids = (NULL_UUID,)
        else:
            transaction_uuids = await transaction.uuids

        while len(transaction_uuids) > 1:
            transaction = await self._transaction_repository.get(uuid=transaction_uuids[-1])
            if transaction.status != TransactionStatus.REJECTED:
                break
            transaction_uuids = tuple(transaction_uuids[:-1])

        return transaction_uuids

    async def _get_delta_entries(self, name: str, uuid: UUID, transaction_uuids: tuple[UUID, ...]) -> list[DeltaEntry]:
        entries = [
            v
            async for v in self._delta_repository.select(name=name, uuid=uuid)
            if v.transaction_uuid in transaction_uuids
        ]

        entries.sort(key=lambda e: (e.version, transaction_uuids.index(e.transaction_uuid)))

        if len({e.transaction_uuid for e in entries}) > 1:
            new = [entries.pop()]
            for e in reversed(entries):
                if e.version < new[-1].version:
                    new.append(e)
            entries = list(reversed(new))
        return entries

    @staticmethod
    def _build_instance(entries: list[DeltaEntry], **kwargs) -> SnapshotEntry:
        if entries[-1].action.is_delete:
            return SnapshotEntry.from_delta_entry(entries[-1])

        cls = entries[0].type_
        instance = cls.from_diff(entries[0].delta, **kwargs)
        for entry in entries[1:]:
            instance.apply_diff(entry.delta)

        snapshot = SnapshotEntry.from_entity(instance)

        return snapshot

    async def _synchronize(self, **kwargs) -> None:
        pass



from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    Iterable,
)
from datetime import (
    datetime,
)
from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
    DatabaseOperationFactory,
)

from ....actions import (
    Action,
)


class DeltaDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Delta Database Operation Factory base class."""

    @abstractmethod
    def build_create(self) -> DatabaseOperation:
        """Build the database operation to create the delta table.

        :return: A ``DatabaseOperation`` instance.s
        """

    @abstractmethod
    def build_submit(
        self,
        transaction_uuids: Iterable[UUID],
        uuid: UUID,
        action: Action,
        name: str,
        version: int,
        data: bytes,
        created_at: datetime,
        transaction_uuid: UUID,
        lock: Optional[int],
        **kwargs,
    ) -> DatabaseOperation:
        """Build the database operation to submit a row into the delta table.

        :param transaction_uuids: The sequence of nested transaction in on top of the current delta's transaction.
        :param uuid: The identifier of the entity.
        :param action: The action of the delta.
        :param name: The name of the entity.
        :param version: The version of the entity
        :param data: The data of the delta.
        :param created_at: The creation datetime.
        :param transaction_uuid: The identifier of the transaction.
        :param lock: The lock identifier.
        :param kwargs: Additional named arguments.
        :return: A ``DatabaseOperation`` instance.
        """

    # noinspection PyShadowingBuiltins
    @abstractmethod
    def build_query(
        self,
        uuid: Optional[UUID] = None,
        name: Optional[str] = None,
        version: Optional[int] = None,
        version_lt: Optional[int] = None,
        version_gt: Optional[int] = None,
        version_le: Optional[int] = None,
        version_ge: Optional[int] = None,
        id: Optional[int] = None,
        id_lt: Optional[int] = None,
        id_gt: Optional[int] = None,
        id_le: Optional[int] = None,
        id_ge: Optional[int] = None,
        transaction_uuid: Optional[UUID] = None,
        transaction_uuid_ne: Optional[UUID] = None,
        transaction_uuid_in: Optional[tuple[UUID, ...]] = None,
        **kwargs,
    ) -> DatabaseOperation:
        """Build the database operation to select rows.

        :param uuid: The identifier must be equal to the given value.
        :param name: The classname must be equal to the given value.
        :param version: The version must be equal to the given value.
        :param version_lt: The version must be lower than the given value.
        :param version_gt: The version must be greater than the given value.
        :param version_le: The version must be lower or equal to the given value.
        :param version_ge: The version must be greater or equal to the given value.
        :param id: The entry identifier must be equal to the given value.
        :param id_lt: The entry identifier must be lower than the given value.
        :param id_gt: The entry identifier must be greater than the given value.
        :param id_le: The entry identifier must be lower or equal to the given value.
        :param id_ge: The entry identifier must be greater or equal to the given value.
        :param transaction_uuid: The transaction identifier must be equal to the given value.
        :param transaction_uuid_ne: The transaction identifier must be distinct of the given value.
        :param transaction_uuid_in: The destination transaction identifier must be equal to one of the given values.

        :return: A ``DatabaseOperation`` instance.
        """

    @abstractmethod
    def build_query_offset(self) -> DatabaseOperation:
        """Build the database operation to get the maximum identifier.

        :return: A ``DatabaseOperation`` instance.
        """



from __future__ import (
    annotations,
)

import logging
from collections import (
    defaultdict,
)
from itertools import (
    chain,
)
from typing import (
    Any,
    Generic,
    Iterable,
    Iterator,
    Type,
    TypeVar,
    Union,
)
from uuid import (
    uuid4,
)

from minos.common import (
    BucketModel,
    Field,
    Model,
    ModelType,
)

from ..actions import (
    Action,
)
from ..collections import (
    IncrementalSet,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


class FieldDiff(Model, Generic[T]):
    """Field Diff class."""

    name: str
    value: T

    def __init__(self, name: str, type_: Type, value: Any, **kwargs):
        super().__init__([self._field_cls("name", str, name), self._field_cls("value", type_, value)], **kwargs)

    @classmethod
    def from_model_type(cls, model_type: ModelType, *args, **kwargs) -> FieldDiff:
        """Build a new instance from model type.

        :param model_type: The model type.
        :param args: Additional positional. arguments.
        :param kwargs: Additional named arguments.
        :return: A new ``FieldDiff`` instance.
        """
        kwargs["type_"] = model_type.type_hints["value"]
        return cls(*args, **kwargs)


class IncrementalFieldDiff(FieldDiff, Generic[T]):
    """Incremental Field Diff class."""

    action: Action

    def __init__(self, name: str, type_: Type, value: Any, action: Action, **kwargs):
        Model.__init__(
            self,
            [
                self._field_cls("name", str, name),
                self._field_cls("value", type_, value),
                self._field_cls("action", Action, action),
            ],
            **kwargs,
        )


class FieldDiffContainer(BucketModel):
    """Field Diff Container class."""

    def __init__(
        self, diffs: Iterable[FieldDiff] = None, fields: Union[Iterable[Field], dict[str, Field]] = None, **kwargs
    ):
        if diffs is not None:
            fields = map(lambda v: self._field_cls(self.generate_random_str(), FieldDiff, v), diffs)

        super().__init__(fields, **kwargs)

        self._mapper = _build_mapper(self._fields)

    def __getitem__(self, item: str) -> Any:
        try:
            return super().__getitem__(item)
        except KeyError as exc:
            try:
                return self.get_one(item)
            except Exception:
                raise exc

    def __eq__(self, other):
        return type(self) == type(other) and tuple(self.values()) == tuple(other.values())

    def __repr__(self) -> str:
        fields_repr = ", ".join(f"{k}={v}" for k, v in self.items())
        return f"{type(self).__name__}({fields_repr})"

    def __iter__(self) -> Iterable[str]:
        """Get the field names.

        :return: An iterable of string values.
        """
        yield from self._mapper.keys()

    def has_one(self, name: str) -> bool:
        """Check if the field identified is present in the container.

        :param name: The name of the field.
        :return: ``True`` if the field is present or ``False`` otherwise.
        """
        return name in self._mapper

    def get_one(self, name: str, return_diff: bool = True) -> Union[FieldDiff, Any, list[FieldDiff], list[Any]]:
        """Get first field diff with given name.

        :param name: The name of the field diff.
        :param return_diff: If ``True`` the result is returned as field diff instances, otherwise the result is
            returned as value instances.
        :return: A ``FieldDiff`` instance.
        """
        type_, names = self._mapper[name]
        if not issubclass(type_, IncrementalFieldDiff):
            return self._get_one(names[0], return_diff)
        return [self._get_one(name, return_diff) for name in names]

    def _get_one(self, name: str, return_diff: bool) -> Union[FieldDiff, Any]:
        diff = getattr(self, name)
        if return_diff:
            return diff
        return diff.value

    def get_all(self, return_diff: bool = True) -> dict[str, Union[FieldDiff, Any, list[FieldDiff], list[Any]]]:
        """Get a dictionary containing all names as keys and all the values of each one as values.

        :param return_diff: If ``True`` the result is returned as field diff instances, otherwise the result is
            returned as value instances.
        :return: A ``dict`` with ``str`` keys and ``list[Any]`` values.
        """
        return {key: self.get_one(key, return_diff) for key in self.keys()}

    def flatten_items(self) -> Iterator[tuple[str, FieldDiff]]:
        """Get the field differences in a flatten way.

        :return: An ``Iterator`` of ``tuple[str, FieldDiff]`` instances.
        """
        return map(lambda diff: (diff.name, diff), self.flatten_values())

    def flatten_values(self) -> Iterator[FieldDiff]:
        """Get the field differences in a flatten way.

        :return: An ``Iterator`` of ``FieldDiff`` instances.
        """
        iterable = (v if isinstance(v, list) else [v] for v in self.values())
        return chain.from_iterable(iterable)

    @classmethod
    def from_difference(cls, a: Model, b: Model, ignore: set[str] = frozenset()) -> FieldDiffContainer:
        """Build a new instance from the difference between two models.

        :param a: Latest model instance.
        :param b: Oldest model instance.
        :param ignore: Set of fields to be ignored.
        :return: A new ``FieldDiffContainer`` instance.
        """
        logger.debug(f"Computing the {cls!r} between {a!r} and {b!r}...")
        differences = cls._diff(a.fields, b.fields)
        differences = [difference for difference in differences if difference.name not in ignore]
        return cls(differences)

    @staticmethod
    def _diff(a: dict[str, Field], b: dict[str, Field]) -> list[FieldDiff]:
        differences = list()
        for a_name, a_field in a.items():
            if a_name not in b or a_field != b[a_name]:
                if isinstance(a_field.value, IncrementalSet):
                    diffs = a_field.value.diff(b[a_name].value).diffs
                    for diff in diffs:
                        differences.append(
                            IncrementalFieldDiff(a_name, a_field.value.data_cls, diff.entity, diff.action)
                        )
                else:
                    differences.append(FieldDiff(a_name, a_field.type, a_field.value))

        return differences

    @classmethod
    def from_model(cls, model: Model, ignore: set[str] = frozenset()) -> FieldDiffContainer:
        """Build a new difference from a single model.

        :param model: The model instance.
        :param ignore: Set of fields to be ignored.
        :return: A new ``FieldDiffContainer`` instance.
        """
        differences = list()
        for field in model.fields.values():
            differences.append(FieldDiff(field.name, field.type, field.value))

        differences = [difference for difference in differences if difference.name not in ignore]
        return cls(differences)

    @staticmethod
    def generate_random_str() -> str:
        """Generate a random string

        :return: A random string value.
        """
        return str(uuid4())


def _build_mapper(fields: dict[str, Field]) -> dict[str, tuple[type, list[str]]]:
    mapper = defaultdict(list)
    for name, field in fields.items():
        mapper[field.value.name].append(name)
    mapper = dict(mapper)

    ans = dict()
    for k, names in mapper.items():
        types = {type(fields[name].value) for name in names}
        if len(types) > 1:
            raise ValueError(f"Multiple {FieldDiff.__name__!r} types have been provided to {k!r}: {types}")
        type_ = next(iter(types))
        if len(names) > 1 and not issubclass(type_, IncrementalFieldDiff):
            raise ValueError(f"Only {IncrementalFieldDiff.__name__!r} type allow multiple {k!r} values.")
        ans[k] = (type_, names)

    return ans



from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    Iterable,
)
from datetime import (
    datetime,
)
from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
    DatabaseOperationFactory,
)

from ....queries import (
    _Condition,
    _Ordering,
)


class SnapshotDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Snapshot Database Operation Factory class."""

    @abstractmethod
    def build_create(self) -> DatabaseOperation:
        """Build the database operation to create the snapshot table.

        :return: A ``DatabaseOperation`` instance.
        """

    @abstractmethod
    def build_delete(self, transaction_uuids: Iterable[UUID]) -> DatabaseOperation:
        """Build the database operation to delete rows by transaction identifiers.

        :param transaction_uuids: The transaction identifiers.
        :return: A ``DatabaseOperation`` instance.
        """

    @abstractmethod
    def build_submit(
        self,
        uuid: UUID,
        name: str,
        version: int,
        schema: bytes,
        data: dict[str, Any],
        created_at: datetime,
        updated_at: datetime,
        transaction_uuid: UUID,
    ) -> DatabaseOperation:
        """Build the insert database operation.

        :param uuid: The identifier of the entity.
        :param name: The name of the entity.
        :param version: The version of the entity.
        :param schema: The schema of the entity.
        :param data: The data of the entity.
        :param created_at: The creation datetime.
        :param updated_at: The last update datetime.
        :param transaction_uuid: The transaction identifier.
        :return: A ``DatabaseOperation`` instance.
        """

    @abstractmethod
    def build_query(
        self,
        name: str,
        condition: _Condition,
        ordering: Optional[_Ordering],
        limit: Optional[int],
        transaction_uuids: tuple[UUID, ...],
        exclude_deleted: bool,
    ) -> DatabaseOperation:
        """Build the query database operation.

        :param name: Class name of the ``Entity``.
        :param condition: The condition that must be satisfied by the ``Entity`` instances.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :param transaction_uuids: The transaction within the operation is performed. If not any value is provided, then
            the transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param exclude_deleted: If ``True``, deleted ``Entity`` entries are included, otherwise deleted
            ``Entity`` entries are filtered.
        :return: A ``DatabaseOperation`` instance.
        """

    @abstractmethod
    def build_submit_offset(self, value: int) -> DatabaseOperation:
        """Build the database operation to store the offset.

        :param value: The value to be stored as the new offset.
        :return: A ``DatabaseOperation`` instance.
        """

    @abstractmethod
    def build_query_offset(self) -> DatabaseOperation:
        """Build the database operation to get the current offset.

        :return: A ``DatabaseOperation`` instance.
        """



import logging
from asyncio import (
    gather,
)
from collections.abc import (
    Iterable,
)
from functools import (
    reduce,
)
from itertools import (
    chain,
)
from operator import (
    or_,
)
from typing import (
    Any,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    Inject,
    Model,
    ModelType,
    NotProvidedException,
    PoolFactory,
)
from minos.networks import (
    BrokerClient,
    BrokerClientPool,
    BrokerMessageV1,
    BrokerMessageV1Payload,
)

from ...entities import (
    Entity,
)
from ...exceptions import (
    RefException,
)
from ...snapshots import (
    SnapshotRepository,
)
from .extractors import (
    RefExtractor,
)
from .injectors import (
    RefInjector,
)

logger = logging.getLogger(__name__)


class RefResolver:
    """Ref Resolver class."""

    # noinspection PyUnusedLocal
    @Inject()
    def __init__(
        self,
        snapshot_repository: SnapshotRepository,
        broker_pool: Optional[BrokerClientPool] = None,
        pool_factory: Optional[PoolFactory] = None,
        **kwargs,
    ):
        if broker_pool is None and pool_factory is not None:
            broker_pool = pool_factory.get_pool("broker")

        if broker_pool is None:
            raise NotProvidedException(f"A {BrokerClientPool!r} instance is required.")

        if snapshot_repository is None:
            raise NotProvidedException(f"A {SnapshotRepository!r} instance is required.")

        self._broker_pool = broker_pool
        self._snapshot_repository = snapshot_repository

    @property
    def broker_pool(self) -> BrokerClientPool:
        """Get the broker pool.

        :return: A ``BrokerClientPool`` instance.
        """
        return self._broker_pool

    @property
    def snapshot_repository(self) -> SnapshotRepository:
        """Get the snapshot repository.

        :return: A ``SnapshotRepository`` instance.
        """
        return self._snapshot_repository

    # noinspection PyUnusedLocal
    async def resolve(self, data: Any, **kwargs) -> Any:
        """Resolve Ref instances.

        :param data: The data to be resolved.
        :param kwargs: Additional named arguments.
        :return: The data instance with model references already resolved.
        """
        missing = RefExtractor(data).build()

        if not len(missing):
            return data

        recovered = await self._query(missing)

        return RefInjector(data, recovered).build()

    async def _query(self, references: dict[type, set[UUID]]) -> dict[UUID, Model]:
        snapshot, broker = dict(), dict()
        for type_, uuids in references.items():
            if issubclass(type_, Entity) or (isinstance(type_, ModelType) and issubclass(type_.model_cls, Entity)):
                snapshot[type_] = uuids
            else:
                broker[type_] = uuids

        parts = await gather(self._query_broker(broker), self._query_snapshot(snapshot))
        ans = reduce(or_, parts)
        return ans

    async def _query_broker(self, references: dict[type, set[UUID]]) -> dict[UUID, Model]:
        if not len(references):
            return dict()

        messages = (
            BrokerMessageV1(self.build_topic_name(type_), BrokerMessageV1Payload({"uuids": refs}))
            for type_, refs in references.items()
        )
        async with self._broker_pool.acquire() as broker:
            futures = (broker.send(message) for message in messages)
            await gather(*futures)

            return {model.uuid: model for model in await self._get_response(broker, len(references))}

    async def _query_snapshot(self, references: dict[type, set[UUID]]) -> dict[UUID, Model]:
        if not len(references):
            return dict()

        futures = list()
        for type_, uuids in references.items():
            for uuid in uuids:
                # noinspection PyTypeChecker
                futures.append(self._snapshot_repository.get(type_, uuid))
        return {model.uuid: model for model in await gather(*futures)}

    @staticmethod
    async def _get_response(broker: BrokerClient, count: int, **kwargs) -> Iterable[Model]:
        messages = list()
        async for message in broker.receive_many(count, **kwargs):
            if not message.ok:
                raise RefException(f"The received message is not ok: {message!r}")
            messages.append(message)

        return chain(*(message.content for message in messages))

    @staticmethod
    def build_topic_name(entity: Union[type, str]) -> str:
        """Build the topic name based on the name of the entity.

        :param entity: The name of the entity to be resolved.
        :return: The topic name.
        """
        if isinstance(entity, type):
            entity = entity.__name__
        return f"_Get{entity}Snapshots"



from __future__ import (
    annotations,
)

import json
from datetime import (
    datetime,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    NULL_UUID,
    MinosJsonBinaryProtocol,
    import_module,
)

from ..deltas import (
    DeltaEntry,
)
from ..exceptions import (
    AlreadyDeletedException,
)

if TYPE_CHECKING:
    from ..entities import (
        Entity,
    )


class SnapshotEntry:
    """Minos Snapshot Entry class.

    Is the python object representation of a row in the ``snapshot`` storage system.
    """

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        uuid: UUID,
        name: str,
        version: int,
        schema: Optional[Union[list[dict[str, Any]], dict[str, Any]], bytes, memoryview] = None,
        data: Optional[Union[dict[str, Any], str]] = None,
        created_at: Optional[datetime] = None,
        updated_at: Optional[datetime] = None,
        transaction_uuid: UUID = NULL_UUID,
    ):
        if isinstance(schema, memoryview):
            schema = schema.tobytes()
        if isinstance(schema, bytes):
            schema = MinosJsonBinaryProtocol.decode(schema)

        if isinstance(data, str):
            data = json.loads(data)

        self.uuid = uuid
        self.name = name
        self.version = version

        self.schema = schema
        self.data = data

        self.created_at = created_at
        self.updated_at = updated_at

        self.transaction_uuid = transaction_uuid

    @classmethod
    def from_entity(cls, instance: Entity, **kwargs) -> SnapshotEntry:
        """Build a new instance from a ``Entity``.

        :param instance: The ``Entity`` instance.
        :return: A new ``SnapshotEntry`` instance.
        """
        data = {k: v for k, v in instance.avro_data.items() if k not in {"uuid", "version", "created_at", "updated_at"}}

        # noinspection PyTypeChecker
        return cls(
            uuid=instance.uuid,
            name=instance.classname,
            version=instance.version,
            schema=instance.avro_schema,
            data=data,
            created_at=instance.created_at,
            updated_at=instance.updated_at,
            **kwargs,
        )

    @classmethod
    def from_delta_entry(cls, entry: DeltaEntry) -> SnapshotEntry:
        """Build a new ``SnapshotEntry`` from a deletion delta.

        :param entry: The delta entry containing the delete information.
        :return: A new ``SnapshotEntry`` instance.
        """
        return cls(
            uuid=entry.uuid,
            name=entry.name,
            version=entry.version,
            created_at=entry.created_at,
            updated_at=entry.created_at,
            transaction_uuid=entry.transaction_uuid,
        )

    def as_raw(self) -> dict[str, Any]:
        """Get a raw representation of the instance.

        :return: A dictionary in which the keys are attribute names and values the attribute contents.
        """
        return {
            "uuid": self.uuid,
            "name": self.name,
            "version": self.version,
            "schema": self.encoded_schema,
            "data": self.encoded_data,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "transaction_uuid": self.transaction_uuid,
        }

    @property
    def encoded_schema(self) -> Optional[bytes]:
        """Get the encoded schema if available.

        :return: A ``bytes`` instance or ``None``.
        """
        if self.schema is None:
            return None

        return MinosJsonBinaryProtocol.encode(self.schema)

    @property
    def encoded_data(self) -> Optional[str]:
        """Get the encoded data if available.

        :return: A ``str`` instance or ``None``.
        """
        if self.data is None:
            return None

        return json.dumps(self.data)

    def build(self, **kwargs) -> Entity:
        """Rebuild the stored ``Entity`` object instance from the internal state.

        :param kwargs: Additional named arguments.
        :return: A ``Entity`` instance.
        """
        from ..entities import (
            Entity,
        )

        if self.data is None:
            raise AlreadyDeletedException(f"The {self.uuid!r} identifier belongs to an already deleted instance.")
        data = dict(self.data)
        data |= {
            "uuid": self.uuid,
            "version": self.version,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }
        data |= kwargs
        instance = Entity.from_avro(self.schema, data)
        return instance

    @property
    def type_(self) -> type[Entity]:
        """Load the concrete ``Entity`` class.

        :return: A ``Type`` object.
        """
        # noinspection PyTypeChecker
        return import_module(self.name)

    def __eq__(self, other: SnapshotEntry) -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __iter__(self) -> Iterable:
        # noinspection PyRedundantParentheses
        yield from (
            self.name,
            self.version,
            self.schema,
            self.data,
            self.created_at,
            self.updated_at,
            self.transaction_uuid,
        )

    def __repr__(self):
        name = type(self).__name__
        return (
            f"{name}(uuid={self.uuid!r}, name={self.name!r}, "
            f"version={self.version!r}, schema={self.schema!r}, data={self.data!r}, "
            f"created_at={self.created_at!r}, updated_at={self.updated_at!r}, "
            f"transaction_uuid={self.transaction_uuid!r})"
        )



from __future__ import (
    annotations,
)

from enum import (
    Enum,
)


class Action(str, Enum):
    """Enum class that describes the actions."""

    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"

    @classmethod
    def value_of(cls, value: str) -> Action:
        """Get the action based on its text representation."""
        for item in cls.__members__.values():
            if item.value == value:
                return item
        raise ValueError(f"The given value does not match with any enum items. Obtained {value}")

    @property
    def is_create(self) -> bool:
        """Check if the action is create.

        :return: A boolean value.
        """
        return self is Action.CREATE

    @property
    def is_update(self) -> bool:
        """Check if the action is create.

        :return: A boolean value.
        """
        return self is Action.UPDATE

    @property
    def is_delete(self) -> bool:
        """Check if the action is create.

        :return: A boolean value.
        """
        return self is Action.DELETE



from __future__ import (
    annotations,
)

from operator import (
    attrgetter,
)
from typing import (
    Any,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
    Union,
    get_args,
)
from uuid import (
    UUID,
)

from minos.common import (
    NULL_UUID,
    DataDecoder,
    DataEncoder,
    DeclarativeModel,
    Model,
    ModelType,
    SchemaEncoder,
)

from ..collections import (
    IncrementalSet,
    IncrementalSetDiff,
)

T = TypeVar("T", bound=Model)


class EntitySet(IncrementalSet[T]):
    """Entity set class."""

    data: dict[str, T]

    def __init__(self, data: Optional[Iterable[T]] = None, *args, **kwargs):
        DeclarativeModel.__init__(self, dict(), *args, **kwargs)

        if isinstance(data, dict):
            iterable = data.values()
        elif data is not None:
            iterable = data
        else:
            iterable = tuple()

        for entity in iterable:
            self.add(entity)

    def add(self, entity: T) -> None:
        """Add an entity.

        :param entity: The entity to be added.
        :return: This method does not return anything.
        """
        if entity.uuid == NULL_UUID:
            raise ValueError(f"The given entity must have a non-null uuid. Obtained entity: {entity!r}")
        self.data[str(entity.uuid)] = entity

    def discard(self, uuid: Union[T, UUID]) -> None:
        """Discard an entity.

        :param uuid: The entity to be discarded.
        :return: This method does not return anything.
        """
        if not isinstance(uuid, UUID):
            uuid = uuid.uuid
        self.data.pop(str(uuid), None)

    def get(self, uuid: UUID) -> T:
        """Get an entity by identifier.

        :param uuid: The identifier of the entity.
        :return: A entity instance.
        """
        return self.data[str(uuid)]

    def __contains__(self, uuid: Union[T, UUID]) -> bool:
        if not isinstance(uuid, UUID):
            if not hasattr(uuid, "uuid"):
                return False
            uuid = uuid.uuid
        return str(uuid) in self.data

    def __iter__(self) -> Iterator[T]:
        yield from self.data.values()

    def __eq__(self, other):
        if isinstance(other, EntitySet):
            return super().__eq__(other)
        if isinstance(other, dict):
            return self.data == other
        return set(self) == other

    def diff(self, another: EntitySet[T]) -> IncrementalSetDiff:
        """Compute the difference between self and another entity set.

        :param another: Another entity set instance.
        :return: The difference between both entity sets.
        """
        return IncrementalSetDiff.from_difference(self, another, get_fn=attrgetter("uuid"))

    @property
    def data_cls(self) -> Optional[type]:
        """Get data class if available.

        :return: A model type.
        """
        args = get_args(self.type_hints["data"])
        return args[1]

    # noinspection PyMethodParameters
    @classmethod
    def encode_schema(cls, encoder: SchemaEncoder, target: Any, **kwargs) -> Any:
        """Encode schema with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded schema.
        :return: The encoded schema of the instance.
        """
        type_ = get_args(target.type_hints["data"])[-1]
        schema = encoder.build(list[type_], **kwargs)
        return schema | {"logicalType": cls.classname}

    @staticmethod
    def encode_data(encoder: DataEncoder, target: Any, **kwargs) -> Any:
        """Encode data with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded data.
        :return: The encoded data of the instance.
        """
        target = list(target["data"].values())
        return encoder.build(target, **kwargs)

    @classmethod
    def decode_data(cls, decoder: DataDecoder, target: Any, type_: ModelType, **kwargs) -> IncrementalSet:
        """Decode data with the given decoder.

        :param decoder: The decoder instance.
        :param target: The data to be decoded.
        :param type_: The data type.
        :return: A decoded instance.
        """
        data_cls = get_args(type_.type_hints["data"])[1]
        target = (decoder.build(v, data_cls, **kwargs) for v in target)
        target = {str(v["uuid"]): v for v in target}
        decoded = decoder.build(target, type_.type_hints["data"], **kwargs)
        return cls(decoded, additional_type_hints=type_.type_hints)



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    Awaitable,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    Injectable,
    ModelType,
    SetupMixin,
    classname,
)
from minos.transactions import (
    TRANSACTION_CONTEXT_VAR,
    TransactionEntry,
)

from ...exceptions import (
    NotFoundException,
)
from ...queries import (
    _TRUE_CONDITION,
    _Condition,
    _EqualCondition,
    _Ordering,
)
from ..entries import (
    SnapshotEntry,
)

if TYPE_CHECKING:
    from ...entities import (
        Entity,
    )


@Injectable("snapshot_repository")
class SnapshotRepository(ABC, SetupMixin):
    """Base Snapshot class.

    The snapshot provides a direct accessor to the ``Entity`` instances stored as deltas by the delta repository
    class.
    """

    async def get(
        self,
        name: Union[str, type[Entity], ModelType],
        uuid: UUID,
        transaction: Optional[TransactionEntry] = None,
        **kwargs,
    ) -> Entity:
        """Get a ``Entity`` instance from its identifier.

        :param name: Class name of the ``Entity``.
        :param uuid: Identifier of the ``Entity``.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param kwargs: Additional named arguments.
        :return: The ``Entity`` instance.
        """
        snapshot_entry = await self.get_entry(name, uuid, transaction=transaction, **kwargs)
        instance = snapshot_entry.build(**kwargs)
        return instance

    async def get_entry(self, name: Union[str, type[Entity], ModelType], uuid: UUID, **kwargs) -> SnapshotEntry:
        """Get a ``SnapshotEntry`` from its identifier.

        :param name: Class name of the ``Entity``.
        :param uuid: Identifier of the ``Entity``.
        :param kwargs: Additional named arguments.
        :return: The ``SnapshotEntry`` instance.
        """

        try:
            return await self.find_entries(
                name, _EqualCondition("uuid", uuid), **kwargs | {"exclude_deleted": False}
            ).__anext__()
        except StopAsyncIteration:
            raise NotFoundException(f"The instance could not be found: {uuid!s}")

    def get_all(
        self,
        name: Union[str, type[Entity], ModelType],
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        streaming_mode: bool = False,
        transaction: Optional[TransactionEntry] = None,
        **kwargs,
    ) -> AsyncIterator[Entity]:
        """Get all ``Entity`` instances.

        :param name: Class name of the ``Entity``.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :param streaming_mode: If ``True`` return the values in streaming directly from the database (keep an open
            database connection), otherwise preloads the full set of values on memory and then retrieves them.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that containing the ``Entity`` instances.
        """
        return self.find(
            name,
            _TRUE_CONDITION,
            ordering=ordering,
            limit=limit,
            streaming_mode=streaming_mode,
            transaction=transaction,
            **kwargs,
        )

    async def find_one(
        self,
        name: Union[str, type[Entity]],
        condition: _Condition,
        transaction: Optional[TransactionEntry] = None,
        **kwargs,
    ) -> Entity:
        """Find a ``Entity`` instance based on a ``Condition``.

        :param name: Class name of the ``Entity``.
        :param condition: The condition that must be satisfied by the ``Entity`` instances.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that containing the ``Entity`` instances.
        """
        try:
            return await self.find(name, condition=condition, limit=1, transaction=transaction, **kwargs).__anext__()
        except StopAsyncIteration:
            raise NotFoundException(f"There are not any instance matching the given condition: {condition}")

    async def find(
        self,
        name: Union[str, type[Entity], ModelType],
        condition: _Condition,
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        streaming_mode: bool = False,
        transaction: Optional[TransactionEntry] = None,
        **kwargs,
    ) -> AsyncIterator[Entity]:
        """Find a collection of ``Entity`` instances based on a ``Condition``.

        :param name: Class name of the ``Entity``.
        :param condition: The condition that must be satisfied by the ``Entity`` instances.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :param streaming_mode: If ``True`` return the values in streaming directly from the database (keep an open
            database connection), otherwise preloads the full set of values on memory and then retrieves them.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that containing the ``Entity`` instances.
        """
        iterable = self.find_entries(
            name=name,
            condition=condition,
            ordering=ordering,
            limit=limit,
            streaming_mode=streaming_mode,
            transaction=transaction,
            **kwargs,
        )
        async for snapshot_entry in iterable:
            yield snapshot_entry.build(**kwargs)

    async def find_entries(
        self,
        name: Union[str, type[Entity], ModelType],
        condition: _Condition,
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        streaming_mode: bool = False,
        transaction: Optional[TransactionEntry] = None,
        exclude_deleted: bool = True,
        synchronize: bool = True,
        **kwargs,
    ) -> AsyncIterator[SnapshotEntry]:
        """Find a collection of ``SnapshotEntry`` instances based on a ``Condition``.

        :param name: Class name of the ``Entity``.
        :param condition: The condition that must be satisfied by the ``Entity`` instances.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :param streaming_mode: If ``True`` return the values in streaming directly from the database (keep an open
            database connection), otherwise preloads the full set of values on memory and then retrieves them.
        :param transaction: The transaction within the operation is performed. If not any value is provided, then the
            transaction is extracted from the context var. If not any transaction is being scoped then the query is
            performed to the global snapshot.
        :param exclude_deleted: If ``True``, deleted ``Entity`` entries are included, otherwise deleted
            ``Entity`` entries are filtered.
        :param synchronize: If ``True`` a synchronization is performed before processing the query, otherwise the query
            is performed without any synchronization step.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that containing the ``Entity`` instances.
        """
        if isinstance(name, ModelType):
            name = name.model_cls
        if isinstance(name, type):
            name = classname(name)

        if transaction is None:
            transaction = TRANSACTION_CONTEXT_VAR.get()

        if synchronize:
            await self.synchronize(**kwargs)

        iterable = self._find_entries(
            name=name,
            condition=condition,
            ordering=ordering,
            limit=limit,
            streaming_mode=streaming_mode,
            transaction=transaction,
            exclude_deleted=exclude_deleted,
            **kwargs,
        )
        async for entry in iterable:
            yield entry

    @abstractmethod
    def _find_entries(
        self,
        name: str,
        condition: _Condition,
        ordering: Optional[_Ordering],
        limit: Optional[int],
        streaming_mode: bool,
        transaction: Optional[TransactionEntry],
        exclude_deleted: bool,
        **kwargs,
    ) -> AsyncIterator[SnapshotEntry]:
        raise NotImplementedError

    def synchronize(self, **kwargs) -> Awaitable[None]:
        """Synchronize the snapshot to the latest available version.

        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        return self._synchronize(**kwargs, synchronize=False)

    @abstractmethod
    async def _synchronize(self, **kwargs) -> None:
        raise NotImplementedError



from typing import (
    Any,
)
from uuid import (
    UUID,
)

from minos.common import (
    Model,
    is_model_type,
)


class RefInjector:
    """Model Reference Injector class."""

    def __init__(self, value: Any, mapper: dict[UUID, Model]):
        self.value = value
        self.mapper = mapper

    def build(self) -> Any:
        """Inject the model instances referenced by identifiers.

        :return: A model in which the model references have been replaced by the values.
        """
        return self._build(self.value)

    def _build(self, value: Any) -> Any:
        if isinstance(value, (tuple, list, set)):
            return type(value)(self._build(v) for v in value)

        if isinstance(value, dict):
            return type(value)((self._build(k), self._build(v)) for k, v in value.items())

        if isinstance(value, UUID) and value in self.mapper:
            return self.mapper[value]

        if is_model_type(value):
            for field in value.fields.values():
                field.value = self._build(field.value)
            return value

        return value



from __future__ import (
    annotations,
)

from datetime import (
    datetime,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Optional,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    NULL_UUID,
    import_module,
)
from minos.transactions import (
    TransactionEntry,
)

from ..actions import (
    Action,
)
from .fields import (
    FieldDiffContainer,
)
from .models import (
    Delta,
)

if TYPE_CHECKING:
    from ..entities import (
        Entity,
    )


class DeltaEntry:
    """Class that represents an entry (or row) on the delta repository database which stores the root entity changes."""

    __slots__ = (
        "uuid",
        "name",
        "version",
        "data",
        "id",
        "action",
        "created_at",
        "transaction_uuid",
    )

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        uuid: UUID,
        name: str,
        version: Optional[int] = None,
        data: Union[bytes, memoryview] = bytes(),
        id: Optional[int] = None,
        action: Optional[Union[str, Action]] = None,
        created_at: Optional[datetime] = None,
        transaction_uuid: UUID = NULL_UUID,
    ):
        if isinstance(data, memoryview):
            data = data.tobytes()
        if action is not None and isinstance(action, str):
            action = Action.value_of(action)

        self.uuid = uuid
        self.name = name
        self.version = version
        self.data = data

        self.id = id
        self.action = action
        self.created_at = created_at
        self.transaction_uuid = transaction_uuid

    @classmethod
    def from_delta(cls, delta: Delta, *, transaction: Optional[TransactionEntry] = None, **kwargs) -> DeltaEntry:
        """Build a new instance from a ``Entity``.

        :param delta: The delta.
        :param transaction: Optional transaction.
        :param kwargs: Additional named arguments.
        :return: A new ``DeltaEntry`` instance.
        """
        if transaction is not None:
            kwargs["transaction_uuid"] = transaction.uuid

        # noinspection PyTypeChecker
        return cls(
            uuid=delta.uuid,
            name=delta.name,
            data=delta.fields_diff.avro_bytes,
            action=delta.action,
            **kwargs,
        )

    @classmethod
    def from_another(cls, another: DeltaEntry, **kwargs) -> DeltaEntry:
        """Build a new instance from another ``DeltaEntry``.

        :param another: The ``DeltaEntry``.
        :param kwargs: Additional named arguments.
        :return: A new ``DeltaEntry`` instance.
        """
        return cls(**(another.as_raw() | kwargs | {"id": None}))

    def as_raw(self) -> dict[str, Any]:
        """Get a raw representation of the instance.

        :return: A dictionary in which the keys are attribute names and values the attribute contents.
        """
        return {
            "uuid": self.uuid,
            "name": self.name,
            "version": self.version,
            "data": self.data,
            "id": self.id,
            "action": self.action.value,
            "created_at": self.created_at,
            "transaction_uuid": self.transaction_uuid,
        }

    @property
    def type_(self) -> type[Entity]:
        """Load the concrete ``Entity`` class.

        :return: A ``Type`` object.
        """
        # noinspection PyTypeChecker
        return import_module(self.name)

    @property
    def delta(self) -> Delta:
        """Get the stored ``Delta`` instance.

        :return: An ``Delta`` instance.
        """
        return Delta(
            self.uuid,
            self.name,
            self.version,
            self.action,
            self.created_at,
            self.field_diff_container,
        )

    @property
    def field_diff_container(self) -> FieldDiffContainer:
        """Get the stored field diff container.

        :return: A ``FieldDiffContainer`` instance.
        """
        if not self.data:
            return FieldDiffContainer.empty()

        return FieldDiffContainer.from_avro_bytes(self.data)

    def __eq__(self, other: "DeltaEntry") -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable:
        yield from (
            self.uuid,
            self.name,
            self.version,
            self.data,
            self.id,
            self.action,
            self.created_at,
            self.transaction_uuid,
        )

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            f"uuid={self.uuid!r}, name={self.name!r}, "
            f"version={self.version!r}, len(data)={len(self.data)!r}, "
            f"id={self.id!r}, action={self.action!r}, created_at={self.created_at!r}, "
            f"transaction_uuid={self.transaction_uuid!r})"
        )



from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
    Union,
    get_args,
)
from uuid import (
    UUID,
    SafeUUID,
)

from minos.common import (
    DataDecoder,
    DataEncoder,
    DeclarativeModel,
    Model,
    ModelType,
    SchemaDecoder,
    SchemaEncoder,
)

from ...contextvars import (
    IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR,
)

logger = logging.getLogger(__name__)
MT = TypeVar("MT", bound=Model)


class Ref(DeclarativeModel, UUID, Generic[MT]):
    """Model Reference."""

    data: Union[MT, UUID]

    @classmethod
    def __class_getitem__(cls, params: Union[type, TypeVar, str]):
        if not isinstance(params, tuple):
            params = (params,)

        def _parse(param: Union[type, TypeVar, str]) -> Union[type, TypeVar]:
            if isinstance(param, str):
                param = ModelType.build(param, uuid=UUID, version=int)
            return param

        params = tuple(_parse(param) for param in params)

        return super().__class_getitem__(params)

    def __init__(self, data: Union[MT, UUID], *args, **kwargs):
        if not isinstance(data, UUID) and not hasattr(data, "uuid"):
            raise ValueError(f"data must be an {UUID!r} instance or have 'uuid' as one of its fields")
        DeclarativeModel.__init__(self, data, *args, **kwargs)

    def __setitem__(self, key: str, value: Any) -> None:
        try:
            return super().__setitem__(key, value)
        except KeyError as exc:
            if key == "uuid":
                self.data = value
                return

            try:
                self.data[key] = value
            except Exception:
                raise exc

    def __getitem__(self, item: str) -> Any:
        try:
            return super().__getitem__(item)
        except KeyError as exc:
            if item == "uuid":
                return self.uuid

            if not self.resolved:
                raise KeyError(f"The reference is not resolved: {self!r}")

            try:
                return self.data[item]
            except Exception:
                raise exc

    def __setattr__(self, key: str, value: Any) -> None:
        try:
            return super().__setattr__(key, value)
        except AttributeError as exc:
            if key == "uuid":
                self.data = value
                return

            try:
                setattr(self.data, key, value)
            except Exception:
                raise exc

    def __getattr__(self, item: str) -> Any:
        try:
            return super().__getattr__(item)
        except AttributeError as exc:
            if item == "data":
                raise exc

            if not self.resolved:
                raise AttributeError(f"The reference is not resolved: {self!r}")

            try:
                return getattr(self.data, item)
            except Exception:
                raise exc

    @property
    def int(self) -> int:
        """Get the UUID as a 128-bit integer.

        :return: An integer value.
        """
        return self.uuid.int

    @property
    def is_safe(self) -> SafeUUID:
        """Get an enum indicating whether the UUID has been generated in a way that is safe.

        :return: A ``SafeUUID`` value.
        """
        return self.uuid.is_safe

    # noinspection PyMethodParameters
    @classmethod
    def encode_schema(cls, encoder: SchemaEncoder, target: Any, **kwargs) -> Any:
        """Encode schema with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded schema.
        :return: The encoded schema of the instance.
        """
        schema = encoder.build(target.type_hints["data"], **kwargs)
        return [(sub | {"logicalType": cls.classname}) for sub in schema]

    @classmethod
    def decode_schema(cls, decoder: SchemaDecoder, target: Any, **kwargs) -> ModelType:
        """Decode schema with the given encoder.

        :param decoder: The decoder instance.
        :param target: The schema to be decoded.
        :return: The decoded schema as a type.
        """
        decoded = decoder.build(target, **kwargs)
        if not isinstance(decoded, ModelType):
            raise ValueError(f"The decoded type is not valid: {decoded}")
        return ModelType.from_model(cls[decoded])

    @staticmethod
    def encode_data(encoder: DataEncoder, target: Any, **kwargs) -> Any:
        """Encode data with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded data.
        :return: The encoded data of the instance.
        """
        target = target["data"]
        if IS_REPOSITORY_SERIALIZATION_CONTEXT_VAR.get() and isinstance(target, dict):
            target = target["uuid"]
        return encoder.build(target, **kwargs)

    @classmethod
    def decode_data(cls, decoder: DataDecoder, target: Any, type_: ModelType, **kwargs) -> Ref:
        """Decode data with the given decoder.

        :param decoder: The decoder instance.
        :param target: The data to be decoded.
        :param type_: The data type.
        :return: A decoded instance.
        """
        decoded = decoder.build(target, type_.type_hints["data"], **kwargs)
        return cls(decoded, additional_type_hints=type_.type_hints)

    def __eq__(self, other):
        return super().__eq__(other) or self.uuid == other or self.data == other

    def __hash__(self):
        return hash(self.uuid)

    @property
    def uuid(self) -> UUID:
        """Get the UUID that identifies the ``Model``.

        :return:
        """
        if not self.resolved:
            return self.data
        return self.data.uuid

    @uuid.setter
    def uuid(self, value: UUID) -> None:
        """Set the uuid that identifies the ``Model``.

        :return: This method does not return anything.
        """
        raise RuntimeError("The 'uuid' must be set through the '__setattr__' method.")  # pragma: no cover

    @property
    def data_cls(self) -> Optional[type]:
        """Get data class if available.

        :return: A model type.
        """
        args = get_args(self.type_hints["data"])
        if args[0] != MT:
            return args[0]
        return None

    # noinspection PyUnusedLocal
    async def resolve(self, force: bool = False, **kwargs) -> None:
        """Resolve the instance.

        :param force: If ``True``, the resolution will be performed also if it is not necessary.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        if not force and self.resolved:
            return

        from .resolvers import (
            RefResolver,
        )

        self.data = await RefResolver(**kwargs).resolve(self)

    @property
    def resolved(self) -> bool:
        """Check if the instance is already resolved.

        :return: ``True`` if resolved or ``False`` otherwise.
        """
        try:
            return not isinstance(self.data, UUID)
        except AttributeError:
            return False

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.data!r})"



from __future__ import (
    annotations,
)

from collections.abc import (
    AsyncIterator,
)
from contextlib import (
    suppress,
)
from typing import (
    TYPE_CHECKING,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    NULL_UUID,
    DatabaseMixin,
    Inject,
    NotProvidedException,
    ProgrammingException,
    import_module,
)
from minos.transactions import (
    TransactionEntry,
    TransactionNotFoundException,
    TransactionRepository,
    TransactionStatus,
)

from ....deltas import (
    Delta,
    DeltaEntry,
    DeltaRepository,
)
from ....exceptions import (
    NotFoundException,
    SnapshotRepositoryConflictException,
)
from ....queries import (
    _Condition,
    _Ordering,
)
from ...entries import (
    SnapshotEntry,
)
from ..abc import (
    SnapshotRepository,
)
from .factories import (
    SnapshotDatabaseOperationFactory,
)

if TYPE_CHECKING:
    from ....entities import (
        Entity,
    )


class DatabaseSnapshotRepository(SnapshotRepository, DatabaseMixin[SnapshotDatabaseOperationFactory]):
    """Database Snapshot Repository class.

    The snapshot provides a direct accessor to the ``Entity`` instances stored as deltas by the delta repository
    class.
    """

    @Inject()
    def __init__(
        self,
        *args,
        delta_repository: DeltaRepository,
        transaction_repository: TransactionRepository,
        database_key: Optional[tuple[str]] = None,
        **kwargs,
    ):
        if database_key is None:
            database_key = ("aggregate", "snapshot")
        super().__init__(*args, database_key=database_key, **kwargs)

        if delta_repository is None:
            raise NotProvidedException("An delta repository instance is required.")

        if transaction_repository is None:
            raise NotProvidedException("A transaction repository instance is required.")

        self._delta_repository = delta_repository
        self._transaction_repository = transaction_repository

    async def _setup(self) -> None:
        await super()._setup()
        operation = self.database_operation_factory.build_create()
        await self.execute_on_database(operation)

    # noinspection PyUnusedLocal
    async def _find_entries(
        self,
        name: str,
        condition: _Condition,
        ordering: Optional[_Ordering],
        limit: Optional[int],
        streaming_mode: bool,
        transaction: Optional[TransactionEntry],
        exclude_deleted: bool,
        **kwargs,
    ) -> AsyncIterator[SnapshotEntry]:
        if transaction is None:
            transaction_uuids = (NULL_UUID,)
        else:
            transaction_uuids = await transaction.uuids

        operation = self.database_operation_factory.build_query(
            name, condition, ordering, limit, transaction_uuids, exclude_deleted
        )

        async for row in self.execute_on_database_and_fetch_all(operation, streaming_mode=streaming_mode):
            yield SnapshotEntry(*row)

    async def is_synced(self, name: str, **kwargs) -> bool:
        """Check if the snapshot has the latest version of a ``Entity`` instance.

        :param name: Class name of the ``Entity`` to be checked.
        :return: ``True`` if it has the latest version for the identifier or ``False`` otherwise.
        """
        offset = await self._load_offset()
        iterable = self._delta_repository.select(id_gt=offset, name=name, **kwargs)
        try:
            await iterable.__anext__()
            return False
        except StopAsyncIteration:
            return True

    async def _synchronize(self, **kwargs) -> None:
        initial_offset = await self._load_offset()

        transaction_uuids = set()
        offset = initial_offset
        async for delta_entry in self._delta_repository.select(id_gt=offset, **kwargs):
            with suppress(SnapshotRepositoryConflictException):
                await self._dispatch_one(delta_entry, **kwargs)

            offset = max(delta_entry.id, offset)
            transaction_uuids.add(delta_entry.transaction_uuid)

        await self._clean_transactions(transaction_uuids)

        await self._store_offset(offset)

    async def _load_offset(self) -> int:
        operation = self.database_operation_factory.build_query_offset()
        # noinspection PyBroadException
        try:
            row = await self.execute_on_database_and_fetch_one(operation)
        except ProgrammingException:
            return 0
        return row[0]

    async def _store_offset(self, offset: int) -> None:
        operation = self.database_operation_factory.build_submit_offset(offset)
        await self.execute_on_database(operation)

    async def _dispatch_one(self, delta_entry: DeltaEntry, **kwargs) -> SnapshotEntry:
        if delta_entry.action.is_delete:
            return await self._submit_delete(delta_entry)

        return await self._submit_update_or_create(delta_entry, **kwargs)

    async def _submit_delete(self, delta_entry: DeltaEntry) -> SnapshotEntry:
        snapshot_entry = SnapshotEntry.from_delta_entry(delta_entry)
        snapshot_entry = await self._submit_entry(snapshot_entry)
        return snapshot_entry

    async def _submit_update_or_create(self, delta_entry: DeltaEntry, **kwargs) -> SnapshotEntry:
        instance = await self._build_instance(delta_entry, **kwargs)

        snapshot_entry = SnapshotEntry.from_entity(instance, transaction_uuid=delta_entry.transaction_uuid)
        snapshot_entry = await self._submit_entry(snapshot_entry)
        return snapshot_entry

    async def _build_instance(self, delta_entry: DeltaEntry, **kwargs) -> Entity:
        diff = delta_entry.delta

        try:
            transaction = await self._transaction_repository.get(uuid=delta_entry.transaction_uuid)
        except TransactionNotFoundException:
            transaction = None

        instance = await self._update_instance_if_exists(diff, transaction=transaction, **kwargs)
        return instance

    async def _update_instance_if_exists(self, delta: Delta, **kwargs) -> Entity:
        # noinspection PyBroadException
        try:
            # noinspection PyTypeChecker
            previous = await self._select_one_instance(delta.name, delta.uuid, **kwargs)
        except NotFoundException:
            # noinspection PyTypeChecker
            cls = import_module(delta.name)
            return cls.from_diff(delta, **kwargs)

        if previous.version >= delta.version:
            raise SnapshotRepositoryConflictException(previous, delta)

        previous.apply_diff(delta)
        return previous

    async def _select_one_instance(self, name: str, uuid: UUID, **kwargs) -> Entity:
        snapshot_entry = await self.get_entry(name, uuid, **kwargs)
        return snapshot_entry.build(**kwargs)

    async def _submit_entry(self, snapshot_entry: SnapshotEntry) -> SnapshotEntry:
        operation = self.database_operation_factory.build_submit(**snapshot_entry.as_raw())
        response = await self.execute_on_database_and_fetch_one(operation)

        snapshot_entry.created_at, snapshot_entry.updated_at = response

        return snapshot_entry

    async def _clean_transactions(self, uuids: set[UUID], **kwargs) -> None:
        if not len(uuids):
            return

        iterable = self._transaction_repository.select(
            uuid_in=uuids,
            status_in=(TransactionStatus.COMMITTED, TransactionStatus.REJECTED),
            **kwargs,
        )
        uuids = {transaction.uuid async for transaction in iterable}

        if not len(uuids):
            return

        operation = self.database_operation_factory.build_delete(uuids)
        await self.execute_on_database(operation)



import logging
from collections.abc import (
    AsyncIterator,
)
from typing import (
    Optional,
    TypeVar,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    NULL_UUID,
    Inject,
    NotProvidedException,
)

from ..actions import (
    Action,
)
from ..deltas import (
    Delta,
    DeltaEntry,
    DeltaRepository,
)
from ..exceptions import (
    DeltaRepositoryException,
)
from ..queries import (
    _Condition,
    _Ordering,
)
from ..snapshots import (
    SnapshotRepository,
)
from .models import (
    Entity,
)

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Entity)


class EntityRepository:
    """Entity Repository class."""

    _delta_repository: DeltaRepository
    _snapshot_repository: SnapshotRepository

    # noinspection PyUnusedLocal
    @Inject()
    def __init__(
        self,
        delta_repository: DeltaRepository,
        snapshot_repository: SnapshotRepository,
        *args,
        **kwargs,
    ):

        if delta_repository is None:
            raise NotProvidedException(f"A {DeltaRepository!r} instance is required.")
        if snapshot_repository is None:
            raise NotProvidedException(f"A {SnapshotRepository!r} instance is required.")

        self._delta_repository = delta_repository
        self._snapshot_repository = snapshot_repository

    async def get(self, type_: type[T], uuid: UUID, **kwargs) -> T:
        """Get one instance from the database based on its identifier.

        :param type_: The of the entity to be looked for.
        :param uuid: The identifier of the instance.
        :return: A ``Entity`` instance.
        """
        # noinspection PyTypeChecker
        return await self._snapshot_repository.get(type_, uuid, **kwargs)

    def get_all(
        self,
        type_: type[T],
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> AsyncIterator[T]:
        """Get all instance from the database.

        :param type_: The of the entity to be looked for.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :return: A ``Entity`` instance.
        """
        # noinspection PyTypeChecker
        return self._snapshot_repository.get_all(type_, ordering, limit, **kwargs)

    async def find_one(self, type_: type[T], condition: _Condition, **kwargs) -> T:
        """Find a ``Entity`` instance based on a ``Condition``.

        :param type_: The type of the entity to be looked for.
        :param condition: The condition that must be satisfied by the ``Entity`` instances.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that containing the ``Entity`` instances.
        """
        return await self._snapshot_repository.find_one(type_, condition, **kwargs)

    def find(
        self,
        type_: type[T],
        condition: _Condition,
        ordering: Optional[_Ordering] = None,
        limit: Optional[int] = None,
        **kwargs,
    ) -> AsyncIterator[T]:
        """Find a collection of instances based on a given ``Condition``.

        :param type_: The type of the entity to be looked for.
        :param condition: The ``Condition`` that must be satisfied by all the instances.
        :param ordering: Optional argument to return the instance with specific ordering strategy. The default behaviour
            is to retrieve them without any order pattern.
        :param limit: Optional argument to return only a subset of instances. The default behaviour is to return all the
            instances that meet the given condition.
        :return: An asynchronous iterator of ``Entity`` instances.
        """
        return self._snapshot_repository.find(type_, condition, ordering, limit, **kwargs)

    async def create(self, type_or_instance: Union[T, type[T]], *args, **kwargs) -> tuple[T, Delta]:
        """Create a new ``Entity`` instance.

        :param type_or_instance: The instance to be created. If it is a ``type`` then the instance is created internally
            using ``args`` and ``kwargs`` as parameters.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A new ``Entity`` instance.
        """
        if "uuid" in kwargs:
            raise DeltaRepositoryException(
                f"The identifier must be computed internally on the repository. Obtained: {kwargs['uuid']}"
            )
        if "version" in kwargs:
            raise DeltaRepositoryException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['version']}"
            )
        if "created_at" in kwargs:
            raise DeltaRepositoryException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['created_at']}"
            )
        if "updated_at" in kwargs:
            raise DeltaRepositoryException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['updated_at']}"
            )
        if isinstance(type_or_instance, type):
            instance: T = type_or_instance(*args, **kwargs)
        else:
            instance = type_or_instance
            if len(args) or len(kwargs):
                raise DeltaRepositoryException(
                    f"Additional parameters are not provided when passing an already built {Entity!r} instance. "
                    f"Obtained: args={args!r}, kwargs={kwargs!r}"
                )

        delta = Delta.from_entity(instance)
        entry = await self._delta_repository.submit(delta)

        self._update_from_repository_entry(instance, entry)

        return instance, entry.delta

    # noinspection PyMethodParameters,PyShadowingBuiltins
    async def update(self, instance: T, **kwargs) -> tuple[T, Optional[Delta]]:
        """Update an existing ``Entity`` instance.

        :param instance: The instance to be updated.
        :param kwargs: Additional named arguments.
        :return: An updated ``Entity``  instance.
        """

        if "version" in kwargs:
            raise DeltaRepositoryException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['version']}"
            )
        if "created_at" in kwargs:
            raise DeltaRepositoryException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['created_at']}"
            )
        if "updated_at" in kwargs:
            raise DeltaRepositoryException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['updated_at']}"
            )

        for key, value in kwargs.items():
            setattr(instance, key, value)

        previous = await self.get(type(instance), instance.uuid)
        delta = instance.diff(previous)
        if not len(delta.fields_diff):
            return instance

        entry = await self._delta_repository.submit(delta)

        self._update_from_repository_entry(instance, entry)

        return instance, entry.delta

    async def save(self, instance: T) -> Optional[Delta]:
        """Store the current instance on the repository.

        If didn't exist previously creates a new one, otherwise updates the existing one.
        """
        is_creation = instance.uuid == NULL_UUID
        if is_creation != (instance.version == 0):
            if is_creation:
                raise DeltaRepositoryException(
                    f"The version must be computed internally on the repository. Obtained: {instance.version}"
                )
            else:
                raise DeltaRepositoryException(
                    f"The uuid must be computed internally on the repository. Obtained: {instance.uuid}"
                )

        values = {
            k: field.value
            for k, field in instance.fields.items()
            if k not in {"uuid", "version", "created_at", "updated_at"}
        }
        if is_creation:
            new, delta = await self.create(type(instance), **values)
            instance._fields |= new.fields
        else:
            _, delta = await self.update(instance, **values)
        return delta

    async def refresh(self, instance: T) -> None:
        """Refresh the state of the given instance.

        :return: This method does not return anything.
        """
        new = await self.get(type(instance), instance.uuid)
        instance._fields |= new.fields

    async def delete(self, instance: T) -> Delta:
        """Delete the given root entity instance.

        :return: This method does not return anything.
        """
        delta = Delta.from_entity(instance, action=Action.DELETE)
        entry = await self._delta_repository.submit(delta)

        self._update_from_repository_entry(instance, entry)
        return entry.delta

    @staticmethod
    def _update_from_repository_entry(instance: T, entry: DeltaEntry) -> None:
        instance.uuid = entry.uuid
        instance.version = entry.version
        if entry.action.is_create:
            instance.created_at = entry.created_at
        instance.updated_at = entry.created_at



from __future__ import (
    annotations,
)

from collections.abc import (
    MutableSet,
)
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
    get_args,
)

from minos.common import (
    DataDecoder,
    DataEncoder,
    DeclarativeModel,
    ModelType,
    SchemaDecoder,
    SchemaEncoder,
)

from .actions import (
    Action,
)

T = TypeVar("T")


class IncrementalSet(DeclarativeModel, MutableSet, Generic[T]):
    """Incremental Set class."""

    data: set[T]

    def __init__(self, data: Optional[Iterable[T]] = None, *args, **kwargs):
        if data is None:
            data = set()
        elif not isinstance(data, set):
            data = {value_obj for value_obj in data}
        super().__init__(data, *args, **kwargs)

    def add(self, value_object: T) -> None:
        """Add a value object.
        :param value_object: The value object to be added.
        :return: This method does not return anything.
        """
        self.data.add(value_object)

    def discard(self, value_object: T) -> None:
        """Remove a value object.
        :param value_object: The value object to be added.
        :return: This method does not return anything.
        """
        self.data.discard(value_object)

    def __contains__(self, value_object: T) -> bool:
        return value_object in self.data

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[T]:
        yield from self.data

    def __eq__(self, other: T) -> bool:
        if isinstance(other, IncrementalSet):
            return super().__eq__(other)
        return set(self) == other

    def diff(self, another: IncrementalSet[T]) -> IncrementalSetDiff:
        """Compute the difference between self and another entity set.
        :param another: Another entity set instance.
        :return: The difference between both entity sets.
        """
        return IncrementalSetDiff.from_difference(self, another)

    @property
    def data_cls(self) -> Optional[type]:
        """Get data class if available.

        :return: A model type.
        """
        args = get_args(self.type_hints["data"])
        return args[0]

    # noinspection PyMethodParameters
    @classmethod
    def encode_schema(cls, encoder: SchemaEncoder, target: Any, **kwargs) -> Any:
        """Encode schema with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded schema.
        :return: The encoded schema of the instance.
        """
        schema = encoder.build(target.type_hints["data"], **kwargs)
        return schema | {"logicalType": cls.classname}

    @classmethod
    def decode_schema(cls, decoder: SchemaDecoder, target: Any, **kwargs) -> ModelType:
        """Decode schema with the given encoder.

        :param decoder: The decoder instance.
        :param target: The schema to be decoded.
        :return: The decoded schema as a type.
        """
        decoded = decoder.build(target, **kwargs)
        return ModelType.from_model(cls[get_args(decoded)[-1]])

    @staticmethod
    def encode_data(encoder: DataEncoder, target: Any, **kwargs) -> Any:
        """Encode data with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded data.
        :return: The encoded data of the instance.
        """
        return encoder.build(target["data"], **kwargs)

    @classmethod
    def decode_data(cls, decoder: DataDecoder, target: Any, type_: ModelType, **kwargs) -> IncrementalSet:
        """Decode data with the given decoder.

        :param decoder: The decoder instance.
        :param target: The data to be decoded.
        :param type_: The data type.
        :return: A decoded instance.
        """
        decoded = decoder.build(target, type_.type_hints["data"], **kwargs)
        return cls(decoded, additional_type_hints=type_.type_hints)


IncrementalSetDiffEntry = ModelType.build("SetDiffEntry", {"action": Action, "entity": Any})


class IncrementalSetDiff(DeclarativeModel):
    """Incremental Set Diff class."""

    diffs: list[IncrementalSetDiffEntry]

    @classmethod
    def from_difference(
        cls, new: IncrementalSet[T], old: IncrementalSet[T], get_fn: Optional[Callable[[T], Any]] = None
    ) -> IncrementalSetDiff:
        """Build a new instance from two entity sets.
        :param new: The new entity set.
        :param old: The old entity set.
        :param get_fn: Optional function to get entries from the set by identifier.
        :return: The difference between new and old.
        """
        differences = cls._diff(new, old, get_fn)
        return cls(differences)

    @staticmethod
    def _diff(new: IncrementalSet[T], old: IncrementalSet[T], get_fn) -> list[IncrementalSetDiffEntry]:
        result = list()
        for value in new - old:
            entry = IncrementalSetDiffEntry(Action.CREATE, value)
            result.append(entry)

        for value in old - new:
            entry = IncrementalSetDiffEntry(Action.DELETE, value)
            result.append(entry)

        if get_fn is not None:
            for value in old & new:
                if value == old.get(value.uuid):
                    continue
                entry = IncrementalSetDiffEntry(Action.UPDATE, value)
                result.append(entry)

        return result



from __future__ import (
    annotations,
)

from typing import (
    Any,
    TypeVar,
)

from minos.common import (
    DeclarativeModel,
    Model,
)

from .collections import (
    IncrementalSet,
)
from .exceptions import (
    ValueObjectException,
)


class ValueObject(DeclarativeModel):
    """Value Object class."""

    def __setitem__(self, key: str, value: Any) -> None:
        raise ValueObjectException("modification of an immutable value object not allowed")


T = TypeVar("T", bound=Model)


class ValueObjectSet(IncrementalSet[T]):
    """Value Object Set class."""

    data: set[T]



from __future__ import (
    annotations,
)

from asyncio import (
    gather,
)
from typing import (
    Generic,
    Optional,
    TypeVar,
    get_args,
)

from minos.common import (
    Config,
    Inject,
    Injectable,
    NotProvidedException,
    SetupMixin,
)
from minos.networks import (
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerPublisher,
)
from minos.saga import (
    SagaManager,
)
from minos.transactions import (
    TransactionRepository,
)

from .actions import (
    Action,
)
from .deltas import (
    Delta,
    DeltaRepository,
    IncrementalFieldDiff,
)
from .entities import (
    Entity,
    EntityRepository,
)
from .snapshots import (
    SnapshotRepository,
)

RT = TypeVar("RT", bound=Entity)


@Injectable("aggregate")
class Aggregate(Generic[RT], SetupMixin):
    """Base Service class"""

    transaction_repository: TransactionRepository
    delta_repository: DeltaRepository
    snapshot_repository: SnapshotRepository
    repository: EntityRepository
    broker_publisher: BrokerPublisher
    saga_manager: SagaManager

    @Inject()
    def __init__(
        self,
        transaction_repository: TransactionRepository,
        delta_repository: DeltaRepository,
        snapshot_repository: SnapshotRepository,
        broker_publisher: BrokerPublisher,
        saga_manager: SagaManager,
        *args,
        **kwargs,
    ):
        if transaction_repository is None:
            raise NotProvidedException(f"A {TransactionRepository!r} object must be provided.")
        if delta_repository is None:
            raise NotProvidedException(f"A {DeltaRepository!r} object must be provided.")
        if snapshot_repository is None:
            raise NotProvidedException(f"A {SnapshotRepository!r} object must be provided.")
        if broker_publisher is None:
            raise NotProvidedException(f"A {BrokerPublisher!r} object must be provided.")
        if saga_manager is None:
            raise NotProvidedException(f"A {SagaManager!r} object must be provided.")

        super().__init__(*args, **kwargs)

        self._check_root()

        self.repository = EntityRepository(delta_repository, snapshot_repository)

        self.transaction_repository = transaction_repository
        self.delta_repository = delta_repository
        self.snapshot_repository = snapshot_repository
        self.broker_publisher = broker_publisher
        self.saga_manager = saga_manager

    @classmethod
    def _from_config(cls, config: Config, **kwargs) -> Aggregate:
        kwargs["broker_publisher"] = cls._get_broker_publisher_from_config(config, **kwargs)
        return cls(**kwargs)

    @staticmethod
    @Inject()
    def _get_broker_publisher_from_config(
        config: Config, broker_publisher: BrokerPublisher, **kwargs
    ) -> BrokerPublisher:
        aggregate_config = config.get_aggregate()

        if "publisher" in aggregate_config:
            publisher_config = aggregate_config["publisher"]
            client_cls = publisher_config.pop("client")
            publisher_kwargs = publisher_config | {"broker_publisher": broker_publisher} | kwargs
            broker_publisher = client_cls.from_config(config, **publisher_kwargs)

        return broker_publisher

    async def _setup(self) -> None:
        await super()._setup()
        await self.broker_publisher.setup()

    async def _destroy(self) -> None:
        await self.broker_publisher.destroy()
        await super()._destroy()

    def _check_root(self):
        self.root  # If root is not valid it will raise an exception.

    @property
    def root(self) -> type[Entity]:
        """Get the root entity of the aggregate.

        :return: A ``Entity`` type.
        """
        # noinspection PyUnresolvedReferences
        bases = self.__orig_bases__
        root = get_args(next((base for base in bases if len(get_args(base))), None))[0]
        if not isinstance(root, type) or not issubclass(root, Entity):
            raise TypeError(f"{type(self)!r} must contain a {Entity!r} as generic value.")
        return root

    async def publish_domain_event(self, delta: Optional[Delta]) -> None:
        """Send a domain delta message containing a delta.

        :param delta: The delta to be sent.
        :return: This method does not return anything.
        """

        if delta is None:
            return

        suffix_mapper = {
            Action.CREATE: "Created",
            Action.UPDATE: "Updated",
            Action.DELETE: "Deleted",
        }
        topic = f"{delta.simplified_name}{suffix_mapper[delta.action]}"
        message = BrokerMessageV1(topic, BrokerMessageV1Payload(delta))
        futures = [self.broker_publisher.send(message)]

        if delta.action == Action.UPDATE:
            for decomposed_delta in delta.decompose():
                diff = next(iter(decomposed_delta.fields_diff.flatten_values()))
                composed_topic = f"{topic}.{diff.name}"
                if isinstance(diff, IncrementalFieldDiff):
                    composed_topic += f".{diff.action.value}"

                message = BrokerMessageV1(composed_topic, BrokerMessageV1Payload(decomposed_delta))
                futures.append(self.broker_publisher.send(message))

        await gather(*futures)



from __future__ import (
    annotations,
)

import logging
import warnings
from datetime import (
    datetime,
)
from functools import (
    total_ordering,
)
from operator import (
    attrgetter,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Union,
)
from uuid import (
    UUID,
)

from minos.common import (
    DeclarativeModel,
)

from ..actions import (
    Action,
)
from .fields import (
    FieldDiff,
    FieldDiffContainer,
)

if TYPE_CHECKING:
    from ..entities import (
        Entity,
    )

logger = logging.getLogger(__name__)


@total_ordering
class Delta(DeclarativeModel):
    """Delta class."""

    uuid: UUID
    name: str
    version: int
    action: Action
    created_at: datetime

    fields_diff: FieldDiffContainer

    @property
    def simplified_name(self) -> str:
        """Get the Entity's simplified name.

        :return: An string value.
        """
        return self.name.rsplit(".", 1)[-1]

    def __lt__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and (
            (self.uuid == other.uuid and self.version < other.version) or (self.created_at < other.created_at)
        )

    def __getitem__(self, item: str) -> Any:
        try:
            return super().__getitem__(item)
        except KeyError as exc:
            if item != "fields_diff":
                try:
                    return self.get_field(item)
                except Exception:
                    raise exc
            raise exc

    def __getattr__(self, item: str) -> Any:
        try:
            return super().__getattr__(item)
        except AttributeError as exc:
            try:
                return self[item]
            except Exception:
                raise exc

    def get_one(self, name: str, return_diff: bool = False) -> Union[FieldDiff, Any, list[FieldDiff], list[Any]]:
        """Get first field diff with given name.

        :param name: The name of the field diff.
        :param return_diff: If ``True`` the result is returned as field diff instances, otherwise the result is
            returned as value instances.
        :return: A ``FieldDiff`` instance.
        """
        warnings.warn("get_one() method is deprecated by get_field() and will be removed soon.", DeprecationWarning)
        return self.get_field(name, return_diff)

    def has_field(self, name: str) -> bool:
        """Check if the field identified is present.

        :param name: The name of the field.
        :return: ``True`` if the field is present or ``False`` otherwise.
        """
        return self.fields_diff.has_one(name)

    def get_field(self, name: str, return_diff: bool = False) -> Union[FieldDiff, Any, list[FieldDiff], list[Any]]:
        """Get first field diff with given name.

        :param name: The name of the field diff.
        :param return_diff: If ``True`` the result is returned as field diff instances, otherwise the result is
            returned as value instances.
        :return: A ``FieldDiff`` instance.
        """
        return self.fields_diff.get_one(name, return_diff)

    def get_all(self, return_diff: bool = False) -> dict[str, Union[FieldDiff, Any, list[FieldDiff], list[Any]]]:
        """Get all field diffs with given name.

        :param return_diff: If ``True`` the result is returned as field diff instances, otherwise the result is
            returned as value instances.
        :return: A list of ``FieldDiff`` instances.
        """
        warnings.warn("get_all() method is deprecated by get_fields() and will be removed soon.", DeprecationWarning)
        return self.get_fields(return_diff)

    def get_fields(self, return_diff: bool = False) -> dict[str, Union[FieldDiff, Any, list[FieldDiff], list[Any]]]:
        """Get all field diffs with given name.

        :param return_diff: If ``True`` the result is returned as field diff instances, otherwise the result is
            returned as value instances.
        :return: A list of ``FieldDiff`` instances.
        """
        return self.fields_diff.get_all(return_diff)

    @classmethod
    def from_difference(cls, a: Entity, b: Entity, action: Action = Action.UPDATE) -> Delta:
        """Build an ``Delta`` instance from the difference of two instances.

        :param a: One ``Entity`` instance.
        :param b: Another ``Entity`` instance.
        :param action: The action that generates the ``Entity`` difference.
        :return: An ``Delta`` instance.
        """
        logger.debug(f"Computing the {cls!r} between {a!r} and {b!r}...")

        if a.uuid != b.uuid:
            raise ValueError(
                f"To compute differences, both arguments must have same identifier. Obtained: {a.uuid!r} vs {b.uuid!r}"
            )

        old, new = sorted([a, b], key=attrgetter("version"))

        fields_diff = FieldDiffContainer.from_difference(a, b, ignore={"uuid", "version", "created_at", "updated_at"})

        return cls(
            uuid=new.uuid,
            name=new.classname,
            version=new.version,
            action=action,
            created_at=new.updated_at,
            fields_diff=fields_diff,
        )

    @classmethod
    def from_entity(cls, instance: Entity, action: Action = Action.CREATE) -> Delta:
        """Build an ``Delta`` from a ``Entity`` (considering all fields as differences).

        :param instance: A ``Entity`` instance.
        :param action: The action that generates the delta.
        :return: An ``Delta`` instance.
        """
        if action != Action.DELETE:
            fields_diff = FieldDiffContainer.from_model(
                instance, ignore={"uuid", "version", "created_at", "updated_at"}
            )
        else:
            fields_diff = FieldDiffContainer.empty()

        return cls(
            uuid=instance.uuid,
            name=instance.classname,
            version=instance.version,
            action=action,
            created_at=instance.updated_at,
            fields_diff=fields_diff,
        )

    def decompose(self) -> list[Delta]:
        """Decompose the ``Delta`` fields into multiple ``Delta`` instances with once Field.

        :return: An list of``Delta`` instances.
        """
        return [
            type(self)(
                uuid=self.uuid,
                name=self.name,
                version=self.version,
                action=self.action,
                created_at=self.created_at,
                fields_diff=FieldDiffContainer([diff]),
            )
            for diff in self.fields_diff.flatten_values()
        ]



from collections import (
    defaultdict,
)
from operator import (
    attrgetter,
)
from typing import (
    Any,
    Iterable,
    Optional,
    Union,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
)

from minos.common import (
    Model,
    TypeHintBuilder,
    is_model_type,
)

from .models import (
    Ref,
)


class RefExtractor:
    """Model Reference Extractor class."""

    def __init__(self, value: Any, type_: Optional[type] = None, as_uuids: bool = True):
        if type_ is None:
            type_ = TypeHintBuilder(value).build()
        self.value = value
        self.type_ = type_
        self.as_uuids = as_uuids

    def build(self) -> dict[type[Model], set[UUID]]:
        """Run the model reference extractor.

        :return: A dictionary in which the keys are the class names and the values are the identifiers.
        """
        ans = defaultdict(set)
        self._build(self.value, self.type_, ans)

        if self.as_uuids:
            ans = {k: set(map(attrgetter("uuid"), v)) for k, v in ans.items()}

        return ans

    def _build(self, value: Any, type_: type, ans: dict[type[Model], set[Ref]]) -> None:
        if get_origin(type_) is Union:
            type_ = next((t for t in get_args(type_) if get_origin(t) is Ref), type_)

        if isinstance(value, (tuple, list, set)):
            self._build_iterable(value, get_args(type_)[0], ans)

        elif isinstance(value, dict):
            self._build_iterable(value.keys(), get_args(type_)[0], ans)
            self._build_iterable(value.values(), get_args(type_)[1], ans)

        elif isinstance(value, Ref):
            cls = value.data_cls
            if cls is None and len(args := get_args(type_)):
                cls = args[0]
            if cls is None and len(args := get_args(type_.type_hints["data"])):
                cls = args[0]
            ans[cls].add(value)

        elif is_model_type(value):
            # noinspection PyUnresolvedReferences
            for field in value.fields.values():
                self._build(field.value, field.type, ans)

    def _build_iterable(self, value: Iterable, type_: type, ans: dict[type[Model], set[Ref]]) -> None:
        for sub_value in value:
            self._build(sub_value, type_, ans)



from datetime import (
    datetime,
)
from typing import (
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DatabaseOperation,
)
from minos.common.testing import (
    MockedDatabaseClient,
    MockedDatabaseOperation,
)

from ....actions import (
    Action,
)
from ....deltas import (
    DeltaDatabaseOperationFactory,
)


class MockedDeltaDatabaseOperationFactory(DeltaDatabaseOperationFactory):
    """For testing purposes."""

    def build_create(self) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("create")

    def build_submit(
        self,
        transaction_uuids: tuple[UUID],
        uuid: UUID,
        action: Action,
        name: str,
        version: int,
        data: bytes,
        created_at: datetime,
        transaction_uuid: UUID,
        lock: Optional[int],
        **kwargs,
    ) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("submit")

    def build_query(
        self,
        uuid: Optional[UUID] = None,
        name: Optional[str] = None,
        version: Optional[int] = None,
        version_lt: Optional[int] = None,
        version_gt: Optional[int] = None,
        version_le: Optional[int] = None,
        version_ge: Optional[int] = None,
        id: Optional[int] = None,
        id_lt: Optional[int] = None,
        id_gt: Optional[int] = None,
        id_le: Optional[int] = None,
        id_ge: Optional[int] = None,
        transaction_uuid: Optional[UUID] = None,
        transaction_uuid_ne: Optional[UUID] = None,
        transaction_uuid_in: Optional[tuple[UUID, ...]] = None,
        **kwargs,
    ) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("select_rows")

    def build_query_offset(self) -> DatabaseOperation:
        """For testing purposes."""
        return MockedDatabaseOperation("select_max_id")


MockedDatabaseClient.set_factory(DeltaDatabaseOperationFactory, MockedDeltaDatabaseOperationFactory)



from .entries import (
    DeltaEntry,
)
from .fields import (
    FieldDiff,
    FieldDiffContainer,
    IncrementalFieldDiff,
)
from .models import (
    Delta,
)
from .repositories import (
    DatabaseDeltaRepository,
    DeltaDatabaseOperationFactory,
    DeltaRepository,
    InMemoryDeltaRepository,
)
